from functools import partial
import traceback
import shutil
from multiprocessing.dummy import Pool as ThreadPool
import os
import sys
import time
import threading
from urllib.request import urlopen, Request
import xml.etree.ElementTree as ET

class GameCache:
	def __init__(self, aId):
		self._dir = ".cache." + str(aId)
		os.makedirs(self._dir, exist_ok=True)
	
	def _GetGameFileName(self, aFileName):
		return os.path.join(self._dir, aFileName)

	def Add(self, aFileName, aData):
		file = open(self._GetGameFileName(aFileName), encoding='utf-8', mode='w')
		file.write(aData)
		file.close()
	
	def Get(self, aFileName):
		if (os.path.exists(self._GetGameFileName(aFileName))):
			with open(self._GetGameFileName(aFileName), encoding='utf-8', mode='r') as myfile:
				return myfile.read()
		return ""
	
	def Destroy(self):
		if (len(self._dir) > 0):
			shutil.rmtree(self._dir)
	
class ProgressReporter:
	def Start(aTotal):
		ProgressReporter.total = aTotal
		ProgressReporter.count = 0
		ProgressReporter.lock = threading.Lock()
		print("Downloading information for " + str(ProgressReporter.total) + " games")
	
	def Update():
		with ProgressReporter.lock:
			ProgressReporter.count += 1
			count = ProgressReporter.count
			p = (100 * count) / ProgressReporter.total 
			print("Progress " + str(count) + " games (" + str(math.ceil(p)) + "%)   ", end='\r', flush=True)


class Utils:
	def GetNum(aString):
		try: 
			return int(aString)
		except:
			return 0
		return 0
	
	def ParseXMLData(root, xmlstring, default):
		try:
			return [x.text for x in root.findall(xmlstring)][0]
		except:
			return default

	def GetYea_lockDate(date):
		try:
			return date.split('/')[2]
		except:
			return ""
		
	def IsAlpha(char):
		if (char>='0' and char<='9'):
			return True
		if (char>='a' and char<='z'):
			return True
		if (char>='A' and char<='Z'):
			return True
		return False

	def GetPureAlphaName(name):
		for char in name:
			if not Utils.IsAlpha(char):
				name = name.replace(char,'')
		return name.lower()

class GamesDB:
	def __init__(self, aCache=None):
		self._cache = aCache
		
	def _API(self, request, params=""):
		try:
			url = "http://thegamesdb.net/api/" + request + ".php"
			if (len(request) > 0):
				url += "?" + params
			s = urlopen(Request(url, headers={'User-Agent': 'Mozilla'}))
			return s.read().decode('utf-8', "ignore")
		except Exception as e:
			if hasattr(e, 'code'):
				print(e.headers)
			else:
				print(e)
		return ""
		
	def GetPlatformsList(self):
		return self._API("GetPlatformsList")
	
	def GetPlatformGames(self, aPlatformId):
		return self._API("GetPlatformGames", "platform=" + str(aPlatformId))

	def GetGame(self, aGameId):
		gameXml = self._cache.Get(aGameId) if (not self._cache is None) else ""
		if (len(gameXml) > 0):
			return gameXml
		
		gameXml = self._API("GetGame", "id=" + aGameId)
		if (not self._cache is None):
			self._cache.Add(aGameId, gameXml)
		return gameXml
		
		
class Platforms:
	def __init__(self):
		xml = GamesDB().GetPlatformsList()
		self._list = [Platform(x) for x in ET.fromstring(xml).findall('*/Platform')]
		self._list = sorted(self._list, key=lambda x: int(x.id))

	def __repr__(self):
		return self._list

	def __str__(self):
		return "Plaforms: " + str(len(self._list))

	def GetList(self):
		return self._list
	
	def Print(self):
		[print(x) for x in self._list] 
		
	def GetPlatformByName(self, aName):
		matches = [x for x in self._list if (x.name == aName)]
		if (len(matches) == 1):
			return matches[0]
		return Platform()

	def GetPlatformById(self, aId):
		matches = [x for x in self._list if (x.id == aId)]
		if (len(matches) == 1):
			return matches[0]
		return Platform()
		
		
class Platform:
	def __init__(self):
		self.id = 0
		self.name = "None"
		self.description = "None"
		
	def __init__(self, root):
		self.id = Utils.GetNum(Utils.ParseXMLData(root, "id", "None"))
		self.name = Utils.ParseXMLData(root, "alias", "None")
		self.description = Utils.ParseXMLData(root, "name", "None")
	
	def __str__(self):
		return str(self.id) + ": " + self.name + ", " + self.description
		
	def __repr__(self):
		return __str__(self)

		
class Game:
	def __init__(self, aGameId, aGamesDB):
		xml = aGamesDB.GetGame(aGameId)
		root = ET.fromstring(xml)

		self.id = Utils.ParseXMLData(root, '*/id', "0")
		self.title = Utils.ParseXMLData(root, '*/GameTitle', "None")
		self.genre = Utils.ParseXMLData(root, '**/genre', "None")
		self.overview = Utils.ParseXMLData(root, '*/Overview', "")
		self.publisher = Utils.ParseXMLData(root, '*/Publisher', "")
		self.players = Utils.ParseXMLData(root, '*/Players', "1")

		date = Utils.ParseXMLData(root, '*/ReleaseDate', "")
		
		self.year = Utils.GetYea_lockDate(date)
		self.name = Utils.GetPureAlphaName(self.title)

	def __str__(self):
		return "Game: " + str(self.id)
		
	def __repr__(self):
		return __str__(self)

	def GetInfo(self):
		buttons = "6";
		simplay = ""
		clone = "";
		ignore = "";
		result = self.name + "=" + self.title + "|" + clone + "|" + self.year + "|" + self.publisher + "|" + self.players + "|" + ignore
		return result + "|" + buttons + "|" + ignore + "|" + simplay + "|" + self.genre

	def GetHistory(self):
		result = "$info=" + self.name + ",\n" + "$bio" + "\n\n" + self.title + " (c) " + self.year + " " + self.publisher + "\n\n"
		return result + self.overview + "\n\n" + "$end" + "\n"


	
class GamesList:
	def __init__(self, aPlatform):
		self.platform = aPlatform
		self._gameCache = GameCache(aPlatform.id)
		self._gamesDB = GamesDB(self._gameCache)
		self._lock = threading.Lock()
		
	def _LoadGame(self, aId):
		game = Game(aId, self._gamesDB)
		self.ReportProgress()
		return game

	def _ReportInitalProgress(self, aTotal):
		with self._lock:
			if (not self._progressReporterFunc is None):
				self._progressReporterFunc(0, aTotal)
	
		
	def Load(self, aProgressReporterFunc=None):
		self._progressReporterFunc = aProgressReporterFunc
		try:
			xml = self._gamesDB.GetPlatformGames(self.platform.id)
			root = ET.fromstring(xml)
			#iDs = [x.text for x in root.findall('*/id')][:25]
			iDs = [x.text for x in root.findall('*/id')]
			
			self.total = len(iDs)
			self.count = 0
			
			self._ReportInitalProgress(self.total)

			pool = ThreadPool()
			gameList = list(pool.map(lambda x: self._LoadGame(x), iDs))
			
			gameList=sorted(gameList, key=lambda x: x.name)
			
			self._gameCache.Destroy()
			
		except Exception as error:
			print("Sorry, an error occured")
			traceback.print_exc()
			gameList = []
			
		return gameList

		
		print("Downloading information for " + str(ProgressReporter.total) + " games")
	
	def ReportProgress(self):
		with self._lock:
			self.count += 1
			if (not self._progressReporterFunc is None):
				self._progressReporterFunc(self.count, self.total)
	

class Output:
	def WriteGameData(aGame, aFunc, aFile):
		aFile.write(aFunc(aGame))
		aFile.write("\n")

	def Save(aFileName, aData, aFunc):
		print("Creating file: " + aFileName)
		file = open(aFileName, encoding='utf-8', mode='w')
		list(map(partial(Output.WriteGameData, aFunc=aFunc, aFile=file), aData))
		file.close()


def CreateOutputFilesForPlatform(aPlatform, aProgressReporterFunc=None):
	try:
		games = GamesList(aPlatform)
		gameList = games.Load(aProgressReporterFunc)

		time.sleep(5)
		Output.Save(aPlatform.name + "_history.dat", gameList, Game.GetHistory)
		Output.Save(aPlatform.name + "_gameinfo.dat", gameList, Game.GetInfo)
	except Exception as error:
		print("Sorry, an error occured")
		traceback.print_exc()





