/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "fileutils.h"
#include "gamedecorator.h"
#include "gamerunner.h"
#include "favsgames.h"
#include "systemmenumodel.h"
#include "systemmenucontroller.h"
#include "sdlapplication.h"
#include "appcommands.h"
#include "gameenginemanager.h"
#include "gamemenus.h"
#include "sdluserproperties.h"
#include "appmenudefs.h"
#include "thememodel.h"
#include "thememenucontroller.h"
#include "themesettingsdecorator.h"
#include "thememenumodel.h"

#include "gameinfocontroller.h"
#include "shutdowncontroller.h"
#include "homecontroller.h"

#include "resourcefileuserproperties.h"
#include "resourcefilejoystickproperties.h"
#include "resourcefilesystemproperties.h"
#include "resourcefilethemeproperties.h"

#include <stdio.h>
#include <string.h>

// needed for WIN32
#undef main

/*****************************************************
 * 
 * GameEngineImageProvider
 * 
 ****************************************************/

class GameEngineImageProvider : public sdl::IImageProvider
{
public:
	GameEngineImageProvider(GameEngineManager& aGameEngineManager) 
		: iGameEngineManager(aGameEngineManager) {}
	
	
	virtual std::string GetFileName(const sdl::MenuItem& aMenuItem, const std::string& aLogicalImageDirectory) override
	{
		return iGameEngineManager.GetMediaFileName(aMenuItem.GetCommand(), aLogicalImageDirectory);
	}
	
	GameEngineManager&	iGameEngineManager;
};


/*****************************************************
 * 
 * Resource files
 * 
 ****************************************************/

ResourceFileSystemProperties	ResourceFileSystemProperties::Instance;
ResourceFileUserProperties		ResourceFileUserProperties::Instance;
ResourceFileJoystickProperties	ResourceFileJoystickProperties::Instance;
ResourceFileThemeProperties		ResourceFileThemeProperties::Instance;

class Configuration
{
public:
	static void AddAppConfig()
	{
		sdl::SDLJoystickProperties.iButtons.push_back(sdl::JoystickProperties::TButton(-1, -1, "AddToFavorites", (int) TAppCommand::EAddtoFavorites));
		sdl::SDLJoystickProperties.iButtons.push_back(sdl::JoystickProperties::TButton(-1, -1, "Game Info", (int) TAppCommand::EGameInfo));
		sdl::SDLJoystickProperties.iButtons.push_back(sdl::JoystickProperties::TButton(-1, -1, "Home", (int) TAppCommand::EHome));
		sdl::SDLJoystickProperties.iButtons.push_back(sdl::JoystickProperties::TButton(-1, -1, "Shutdown", (int) TAppCommand::EShutdown));
	}
	
	static void Load(const std::string& aSystemFileName)
	{
		AddAppConfig();
		
		ResourceFileSystemProperties::Instance.AddResources(sdl::SDLSystemProperties, AppSystemProperties);
		ResourceFileUserProperties::Instance.AddResources(AppUserProperties, sdl::SDLUserProperties);
		ResourceFileJoystickProperties::Instance.AddResources(sdl::SDLJoystickProperties);
		ResourceFileThemeProperties::Instance.AddResources(sdl::SDLThemeProperties, AppThemeProperties);
		
		ResourceFileSystemProperties::Instance.Load(aSystemFileName, ".");
		ResourceFileUserProperties::Instance.Load("config.rc", AppSystemProperties.iUserDir);
		ResourceFileJoystickProperties::Instance.Load("joystick.rc", AppSystemProperties.iUserDir);
		ResourceFileThemeProperties::Instance.Load(ThemeModel::GetThemeFileName(), ThemeModel::GetCurrentThemeDirName());
		
		AppSystemProperties.iImageDirs.push_back(ThemeModel::GetCurrentThemeDirName());
		
		SanityCheck();
	}
	
	static void SanityCheckSystemDirectories()
	{
		if (!FileUtils::DirectoryExists(AppSystemProperties.iUserDir))
		{
			std::cout << "Fatal error: user directory \"" << AppSystemProperties.iUserDir << "\" does not exist" << std::endl;
			exit(-1);
		}
		
		if (!FileUtils::DirectoryExists(AppSystemProperties.iDataDir))
		{
			std::cout << "Fatal: data directory \"" << AppSystemProperties.iDataDir << "\" does not exist" << std::endl;
			exit(-1);
		}
		
		if (!FileUtils::DirectoryExists(AppSystemProperties.iThemeDir))
		{
			std::cout << "Fatal error: theme directory \"" << AppSystemProperties.iThemeDir << "\" does not exist" << std::endl;
			exit(-1);
		}
		
		for (auto &dir : AppSystemProperties.iFontDirs)
		{
			if (!FileUtils::DirectoryExists(dir))
				std::cout << "Warning: font directory \"" << dir << "\" does not exist" << std::endl;
		}
		
		for (auto &dir : AppSystemProperties.iImageDirs)
		{
			if (!FileUtils::DirectoryExists(dir))
				std::cout << "Warning: image directory \"" << dir << "\" does not exist" << std::endl;
		}
	}
	
	static void SanityCheckConfig()
	{
		if (AppUserProperties.iFilterGames.iMaxPlayers > 4)
			std::cout << "Warning: FilterGamesMaxPlayers \"" << AppUserProperties.iFilterGames.iMaxPlayers << "\" cannot be higher than 4 - set to not filter on maximum number of players" << std::endl;
		else  if (AppUserProperties.iFilterGames.iMaxPlayers < 0)
			AppUserProperties.iFilterGames.iMaxPlayers = -1;
		if (AppUserProperties.iFilterGames.iMaxButtons > 16)
			std::cout << "Warning: FilterGamesMaxButtons \"" << AppUserProperties.iFilterGames.iMaxButtons << "\" cannot be higher than 16 - set to not filter on maximum number of buttons" << std::endl;
		else if (AppUserProperties.iFilterGames.iMaxButtons <= 0)
			AppUserProperties.iFilterGames.iMaxButtons = -1;
		
		if ((sdl::SDLUserProperties.iAudioProperties.iVolume < 0) || (sdl::SDLUserProperties.iAudioProperties.iVolume > 100))
		{
			std::cout << "Warning: AppUserProperties.Audio.Volume (" << sdl::SDLUserProperties.iAudioProperties.iVolume << ") should be in range 0-100. Set to 50." << std::endl;
			sdl::SDLUserProperties.iAudioProperties.iVolume = 50;
		}
	}
	
	static void SanityCheckGameEngines()
	{
		for (auto &gameEngineProperties : AppSystemProperties.iGameEnginesProperties)
		{
			if (!FileUtils::DirectoryExists(gameEngineProperties.second.iRomDirectory.iPathName))
				std::cout << "Warning: " + gameEngineProperties.first + " rom directory \"" << gameEngineProperties.second.iRomDirectory.iPathName << "\" does not exist" << std::endl;
			
			if (!FileUtils::FileExists(gameEngineProperties.second.iExeName))
				std::cout << "Warning: " + gameEngineProperties.first + " executable file \"" << gameEngineProperties.second.iExeName << "\" does not exist" << std::endl;
			
			for (int i=0 ; i<KNumOfUserMediaDirectories ; i++)
			{
				auto& dir = gameEngineProperties.second.iMediaDirectories[i];
				if ((dir.iPathName.size() > 0) && (!FileUtils::DirectoryExists(dir.iPathName)))
					std::cout << "Warning: for engine \"" << gameEngineProperties.first << "\", directory \"" << dir.iPathName << "\" does not exist" << std::endl;
			}
		}
	}
	
	static void SanityCheck()
	{
		SanityCheckSystemDirectories();
		SanityCheckConfig();
		SanityCheckGameEngines();
	}
};

/*****************************************************
 * 
 * RGLApplication
 * 
 ****************************************************/

class RGLApplication : public sdl::Application
{
public:
	RGLApplication() : 
		sdl::Application("RGL"),
		iGameEngineImageProvider(iGameEngineManager),
		iGameRunner(iGameEngineManager),
		iThemeMenuModel(iThemeModel),
		iSystemMenuModel(iThemeMenuModel),
		iGameDecorator(iGameEngineManager), 
		iMenu2dWidget(iMenuModel, iGameEngineImageProvider),
		iThemeSettingsDecorator(iThemeModel),
		iGameMenus(iMenuModel, iGameEngineManager),
		iThemeMenuController(iThemeModel), 
		iGameInfoController(iGameEngineManager),
		iSystemMenuController(iGameMenus, GetAudio())
		{}
	~RGLApplication();
	virtual void Construct() override;
	virtual void Run() override;

	bool WantQuit();
protected:
	void BaseConstruct();
	void ConstructEngines();
	void ConstructMenus();
	void ConstructUI();
	
	void AddMenuListeners();
	void AddShortcuts();
	void SetHighlightedMenuItem();
	void SaveHighlightedMenuItem();
	bool iWantQuit = true;

	GameEngineManager 		iGameEngineManager;
	GameEngineImageProvider	iGameEngineImageProvider;
	sdl::Menu2DModel 		iMenuModel;
	ThemeModel 				iThemeModel;
	FavsGames::Model 		iFavsGamesModel;
	GameRunner::Model 		iGameRunner;
	ThemeMenuModel			iThemeMenuModel;
	SystemMenuModel			iSystemMenuModel;
	
	GameDecorator 			iGameDecorator;
	sdl::Menu2DWidget 		iMenu2dWidget;
	ThemeSettingsDecorator	iThemeSettingsDecorator;
	
	GameMenus 				iGameMenus;
	ThemeMenuController		iThemeMenuController;
	GameInfoController 		iGameInfoController;
	SystemMenuController	iSystemMenuController;
	ShutdownController		iShutdownController;
	HomeController			iHomeController = HomeController(iGameMenus, iMenu2dWidget);
};

RGLApplication::~RGLApplication()
{
	AppThemeProperties.SetDefaults();
	AppSystemProperties.SetDefaults();
};

void RGLApplication::Construct()
{
	sdl::System::FontDirs.SetDirectoryNames(AppSystemProperties.iFontDirs);
	sdl::System::ImageDirs.SetDirectoryNames(AppSystemProperties.iImageDirs);
	sdl::System::AudioDirs.SetDirectoryNames(AppSystemProperties.iAudioDirs);
	
	sdl::Application::Construct();

	BaseConstruct();
	ConstructEngines();
	ConstructMenus();
	ConstructUI();
	AddMenuListeners();
}

void RGLApplication::BaseConstruct()
{
	// only after sdl::Application::Construct
	AddShortcuts();
}

void RGLApplication::AddShortcuts()
{
	sdl::EventManager::AddKeyTranslation(SDLK_f, true, false, (int) TAppCommand::EAddtoFavorites);
	sdl::EventManager::AddKeyTranslation(SDLK_i, true, false, (int) TAppCommand::EGameInfo);
	sdl::EventManager::AddKeyTranslation(SDLK_s, true, false, (int) TAppCommand::EShutdown);
	sdl::EventManager::AddKeyTranslation(SDLK_HOME, true, false, (int) TAppCommand::EHome);
	sdl::EventManager::AddKeyTranslation(SDLK_q, true, false, (int) sdl::TCommand::EQuit);
}


bool RGLApplication::WantQuit()
{
	return iWantQuit;
}

void RGLApplication::ConstructEngines()
{
	iGameEngineManager.SetMaxPlayers(AppUserProperties.iFilterGames.iMaxPlayers);
	iGameEngineManager.SetMaxButtons(AppUserProperties.iFilterGames.iMaxButtons);
	iGameEngineManager.SetSimultaneousPlayers(AppUserProperties.iFilterGames.iSimultaneousPlayers);
	iGameEngineManager.Construct();
	
	iGameRunner.Construct();
	iFavsGamesModel.Construct();
	iThemeMenuModel.Construct();
	iSystemMenuModel.Construct();
}

void RGLApplication::ConstructMenus()
{
	iGameMenus.Construct();
	iMenuModel.AddMenu(iSystemMenuModel.GetMenu());	// adds "System" menu
	iMenuModel.AddMenu(iGameRunner.GetMenu());		// adds "Recent games" menu
	iMenuModel.AddMenu(iFavsGamesModel.GetMenu());	// adds "Favorite games" menu

	SetHighlightedMenuItem();
}

void RGLApplication::SetHighlightedMenuItem()
{
	std::string themeName;
	if (ThemeModel::TestAndResetHasThemeChangedLastRun(themeName))
		iMenuModel.HighlightMenuItem(KMenuNameTheme, themeName);
	else
		iMenuModel.HighlightMenuItem(AppUserProperties.iLastMenuName, AppUserProperties.iLastMenuItemName);
}

void RGLApplication::SaveHighlightedMenuItem()
{
	iMenuModel.GetHighlightMenuItem(AppUserProperties.iLastMenuName, AppUserProperties.iLastMenuItemName);
	ResourceFileUserProperties::Instance.Save(); // no user feedback and ignore errors
}


void RGLApplication::ConstructUI()
{
	GetApplicationWidget().SetBackgroundImage(AppThemeProperties.iBackgroundImage);

	iGameDecorator.Construct();
	iMenu2dWidget.AddImageMenuType(KMenuTypeGame); // call before construct
	iMenu2dWidget.AddImageMenuType(KMenuTypeRecent); // call before construct
	iMenu2dWidget.AddImageMenuType(KMenuTypeFavs); // call before construct
	iMenu2dWidget.Construct();
	iThemeSettingsDecorator.Construct();
	
	GetApplicationWidget().AddWidget(iMenu2dWidget);
	GetApplicationWidget().AddWidget(iGameDecorator);
	GetApplicationWidget().AddWidget(iThemeSettingsDecorator);
}

void RGLApplication::AddMenuListeners()
{
	iMenuModel.AddMenuChangedListener(iGameDecorator);
	iMenuModel.AddMenuChangedListener(iThemeSettingsDecorator);
	
	iMenuModel.AddMenuItemListener(iGameDecorator, KMenuTypeGame);
	iMenuModel.AddMenuItemListener(iGameDecorator, KMenuTypeFavs);
	iMenuModel.AddMenuItemListener(iGameDecorator, KMenuTypeRecent);
	
	iMenuModel.AddMenuItemListener(iGameRunner.GetRecentGamesMenuController(), KMenuTypeGame);
	iMenuModel.AddMenuItemListener(iGameRunner.GetRecentGamesMenuController(), KMenuTypeFavs);
	iMenuModel.AddMenuItemListener(iGameRunner.GetRecentGamesMenuController(), KMenuTypeRecent);
	
	iMenuModel.AddMenuItemListener(iGameRunner.GetSettingMenuController(), KMenuTypeSettingsSystemMenus);
	iMenuModel.AddMenuItemListener(iFavsGamesModel.GetFavsGamesMenuController(), KMenuTypeGame);
	iMenuModel.AddMenuItemListener(iFavsGamesModel.GetFavsGamesMenuController(), KMenuTypeFavs);
	iMenuModel.AddMenuItemListener(iFavsGamesModel.GetFavsGamesMenuController(), KMenuTypeRecent);
	iMenuModel.AddMenuItemListener(iFavsGamesModel.GetSettingsMenuController(), KMenuTypeSettingsSystemMenus);
	iMenuModel.AddMenuItemListener(iThemeSettingsDecorator, KMenuTypeSettingsThemeMenu);
	iMenuModel.AddMenuItemListener(iGameInfoController, KMenuTypeGame);
	iMenuModel.AddMenuItemListener(iGameInfoController, KMenuTypeFavs);
	iMenuModel.AddMenuItemListener(iGameInfoController, KMenuTypeRecent);
	iMenuModel.AddMenuItemListener(iThemeMenuController, KMenuTypeSettingsThemeMenu);
	
	iMenuModel.AddMenuItemListener(iSystemMenuController, KMenuTypeSettingsSystemMenus);
	iMenuModel.AddMenuItemListener(iSystemMenuController, KMenuTypeSystem);
	iMenuModel.AddMenuItemListener(iShutdownController);
	iMenuModel.AddMenuItemListener(iHomeController);
}

void RGLApplication::Run()
{
	// In order to launch emulators or game ports that are SDL based, we have to close our SDL
	// SDL resources. That's why we need to come out of sdl::Application::Run in order to
	// launch the game.
	// Similarly, for theme switching we need to go back to DoMain to re-load the config files
	bool quit = true;
	do
	{
		sdl::Application::Run();
		
		quit = true;
		
		if(iGameRunner.WantRunGame())
		{
			bool success = iGameRunner.RunGame();
			if (!success)
				sdl::InfoPrint::Message("Could not run game");
			
			quit = false;
		}
	} while (quit == false);
	
	SaveHighlightedMenuItem();
	
	iWantQuit = !iThemeModel.ThemeChanged();
}


/*****************************************************
 * 
 * Main
 * 
 ****************************************************/

bool DoMain(int argc, char* argv[])
{
	Configuration::Load(std::string((argc == 2) ? argv[1] : Platform::GetPlatformName() + "_system.rc"));
	
	RGLApplication application;
	application.Construct();
	application.Run();
	
	return application.WantQuit();
}

int main(int argc, char* argv[])
{
	bool quit = true;
	do 
	{
		quit = DoMain(argc, argv);
		
	} while(!quit);
	
	return 0;
}
	