/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "gamedecorator.h"
#include "fileutils.h"
#include "stringutils.h"
#include "systemproperties.h"
#include "userproperties.h"
#include "appmenudefs.h"
#include "gameenginemanager.h"

#include "platform.h"

static std::string GetSubstitutedText(const std::string& aText, const StringUtils::TSubstituteMap& aSubstituteMap)
{
	std::string text(aText);
	StringUtils::Substitute(text, aSubstituteMap);

	return text;
}

static void BuildCategorySubstituteMap(StringUtils::TSubstituteMap& aSubstituteMap, const std::string& aCategory, int aNumItems)
{
	aSubstituteMap["%cat%"] = aCategory;
	aSubstituteMap["%numitems%"] = std::to_string(aNumItems);
}

static std::string GetPlayerControl(const TGame& aGame)
{
	if (aGame.iPlayers > 0)
		return std::to_string(aGame.iPlayers) + "P " + (aGame.iSimulatenousPlay ? "sim" : aGame.iAlternatePlay ? "alt" : "");
	return "???";
}

static void BuildGameSubstituteMap(StringUtils::TSubstituteMap& aSubstituteMap, const TGame& aGame)
{
	aSubstituteMap["%name%"]	= aGame.iName;
	aSubstituteMap["%category%"] = aGame.iCategory;
	aSubstituteMap["%year%"] 	= (aGame.iYear.size() > 0) ? aGame.iYear : "???";
	aSubstituteMap["%man%"] 	= (aGame.iManufacturer.size() > 0) ? aGame.iManufacturer : "???";
	aSubstituteMap["%playcon%"] = GetPlayerControl(aGame);
	aSubstituteMap["%buttons%"] = (aGame.iButtons > 0) ? std::to_string(aGame.iButtons) : "?";
	aSubstituteMap["%game%"] 	= (aGame.iDescription.size() > 0) ? aGame.iDescription : "???";
	aSubstituteMap["%players%"] = (aGame.iPlayers > 0) ? std::to_string(aGame.iPlayers) : "?";
	aSubstituteMap["%clone%"] 	= (aGame.iCloneOf.size() > 0) ? aGame.iCloneOf : "???";
	aSubstituteMap["%engine%"] 	= aGame.iEngine;
}

static void UpdateSimpleText(sdl::SimpleTextWidget& aSimpleTextWidget, const ThemeProperties::SimpleTextProperties& aSimpleTextResource, const std::string& aText)
{
	aSimpleTextWidget.SetText(aText);
	aSimpleTextWidget.SetWidth(aSimpleTextResource.iRect.Width());
	aSimpleTextWidget.Layout();
}


void GameDecorator::UpdateGameImage(sdl::ImageWidget& aImage, const ThemeProperties::GameImageProperties& aGameImageResource, const TGame* aGame)
{
	std::string fileName = aGameImageResource.iFileName;
	
	if (aGameImageResource.iFileName.size() > 0)
	{
		if (aGame)
		{
			StringUtils::TSubstituteMap substituteMap;
			BuildGameSubstituteMap(substituteMap, *aGame);
			fileName = GetSubstitutedText(aGameImageResource.iFileName, substituteMap);
		}
		
		// files that still have a % should have been name substituted - we ignore them
		if (fileName.find('%') != std::string::npos)
			fileName.clear();
	}
	else if (aGame)
	{
		fileName = iGameEngineManager.GetMediaFileName(*aGame, aGameImageResource.iLogicalDirectory);
		if (!FileUtils::FileExists(fileName))
			fileName = aGameImageResource.iDefaultImage;
	}
	
	aImage.SetImageFile(fileName);
	aImage.SetEnabled(true);
}


GameDecorator::~GameDecorator()
{
	CloseTimer();
}

void GameDecorator::ConstructGameTexts()
{
	for (int i = 0; i<KNumOfSimpleTexts; i++)
	{
		if (!AppThemeProperties.iGameSimpleText[i].iRect.IsEmpty())
		{
			iText[i].SetName(std::string("GameText: ") + std::to_string(i));
			iText[i].OverrideDefaults(AppThemeProperties.iGameSimpleText[i].iProperties);
			iText[i].SetText(AppThemeProperties.iGameSimpleText[i].iText);
			iText[i].SetRect(AppThemeProperties.iGameSimpleText[i].iRect);
			iText[i].SetShowDebugOutline();
			AddWidget(iText[i]);
		}
	}
}

void GameDecorator::ConstructGameVideo()
{
	sdl::TRect rect = AppThemeProperties.iGameVideoProperties.iRect.IsEmpty() ? AppThemeProperties.iGameImageProperties[0].iRect : AppThemeProperties.iGameVideoProperties.iRect;
	iVideoWidget.SetRect(rect);
	AddWidget(iVideoWidget);
}

void GameDecorator::ConstructGameImages()
{
	for (int i = 0; i<KNumOfGameImages; i++)
	{
		if (!AppThemeProperties.iGameImageProperties[i].iRect.IsEmpty())
		{
			iGameImage[i].SetName(std::string("GameImage: ") + std::to_string(i));
			iGameImage[i].OverrideDefaults(AppThemeProperties.iGameImageProperties[i]);
			iGameImage[i].SetShowDebugOutline();
			AddWidget(iGameImage[i]);
		}
	}
}

void GameDecorator::ConstructCategoryLabel()
{
	iCategoryText.SetName("CategoryText");
	iCategoryText.OverrideDefaults(AppThemeProperties.iCategoryTextProperties.iProperties);
	iCategoryText.SetText(AppThemeProperties.iCategoryTextProperties.iText);
	iCategoryText.SetRect(AppThemeProperties.iCategoryTextProperties.iRect);
	iCategoryText.SetShowDebugOutline();
	AddWidget(iCategoryText);
}


void GameDecorator::Construct()
{
	SetRect(sdl::TRect(0, 0, sdl::SDLThemeProperties.iScreen.iWidth, sdl::SDLThemeProperties.iScreen.iHeight));

	ConstructGameImages();
	ConstructGameTexts();
	ConstructCategoryLabel();
	ConstructGameVideo(); // after ConstructGameImages sothat it's displayed on top
	
	Layout();
}


void GameDecorator::UpdateGameInfo(const TGame* aGame)
{
	iVideoWidget.Stop();


	for (int i=0 ; i< KNumOfSimpleTexts ; i++)
	{
		std::string text = "";
		if (aGame)
		{
			StringUtils::TSubstituteMap substituteMap;
			BuildGameSubstituteMap(substituteMap, *aGame);
			text = GetSubstitutedText(AppThemeProperties.iGameSimpleText[i].iText, substituteMap);
		}
		UpdateSimpleText(iText[i], AppThemeProperties.iGameSimpleText[i], text);
	}
	
	for (int i=0 ; i< KNumOfGameImages ; i++)
	{
		if (!AppThemeProperties.iGameImageProperties[i].iRect.IsEmpty())
			UpdateGameImage(iGameImage[i], AppThemeProperties.iGameImageProperties[i], aGame);
	}

	StartTimer();
}

void GameDecorator::MenuItemHighlighted(const sdl::MenuItem& aMenuItem)
{
	if (iCurrentGameID != aMenuItem.GetCommand())
	{
		iCurrentGameID = aMenuItem.GetCommand();
		UpdateGameInfo(iGameEngineManager.GetGame(aMenuItem.GetCommand()));
		sdl::System::RequestReRender();
	}
}

void GameDecorator::SetCatergoryText(const std::string& aMenuName, int aNumberOfItems)
{
	std::string text;
	if (iIsGameMenu)
	{
		StringUtils::TSubstituteMap substituteMap;
		BuildCategorySubstituteMap(substituteMap, aMenuName, aNumberOfItems);
		text = GetSubstitutedText(AppThemeProperties.iCategoryTextProperties.iText, substituteMap);
	}
	else
		text = AppThemeProperties.iSystemMenuCategoryText + aMenuName;

	UpdateSimpleText(iCategoryText, AppThemeProperties.iCategoryTextProperties, text);
}

void GameDecorator::MenuListChanged(sdl::Menu& aMenu)
{
	SetCatergoryText(aMenu.GetName(), aMenu.MenuSize());
	UpdateGameInfo(nullptr);
}

void GameDecorator::MenuChanged(sdl::Menu& aMenu)
{
	iIsGameMenu = (aMenu.GetType().compare(KMenuTypeGame) == 0) || (aMenu.GetType().compare(KMenuTypeFavs) == 0) || (aMenu.GetType().compare(KMenuTypeRecent) == 0);
	SetCatergoryText(aMenu.GetName(), aMenu.MenuSize());
//	ClearGameInfo();
	iCurrentGameID.clear();
	UpdateGameInfo(nullptr);
}

void GameDecorator::PlayVideo()
{
	std::string videoFileName = iGameEngineManager.GetMediaFileName(iCurrentGameID, AppThemeProperties.iGameVideoProperties.iLogicalDirectory);
	
	iVideoWidget.ShowTimeLabel(AppUserProperties.iVideoSetting.iTimeLabel);
	if (iVideoWidget.PlayVideoFile(videoFileName))
		iGameImage[0].SetEnabled(false);
}

void GameDecorator::Render(sdl::IRenderer& aRenderer)
{
	if (iPlayVideo)
	{
		PlayVideo();
		iPlayVideo = false;
	}
	
	if (!iVideoWidget.IsPlaying())
		iGameImage[0].SetEnabled(true);
	
	sdl::Container::Render(aRenderer);
}


void GameDecorator::CloseTimer()
{
	if (iTimerId != 0)
	{
		SDL_RemoveTimer(iTimerId);
		iTimerId = 0;
	}
	iPlayVideo = false;
}

void GameDecorator::StartTimer()
{
	CloseTimer();
	if (AppUserProperties.iVideoSetting.iVideoEnabled)
		iTimerId = SDL_AddTimer(AppUserProperties.iVideoSetting.iStartAfterDelay, TimerCallback, this);
}

Uint32 GameDecorator::TimerCallback(Uint32 aInterval, void* aParam)
{
	GameDecorator* gameDecorator = (GameDecorator*) aParam;
	return gameDecorator->TimerCallback(aInterval);
}

Uint32 GameDecorator::TimerCallback(Uint32 aInterval)
{
	iPlayVideo = true;
	sdl::EventManager::InjectRedrawEvent();
	return 0;
}

void GameDecorator::ResetSDLState()
{
	Container::ResetSDLState();
	iVideoWidget.Stop();
	iPlayVideo = false;
}
