/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "gameengine.h"
#include "fileutils.h"
#include "platform.h"
#include "stringutils.h"
#include <string.h>
#include <utility>

const std::string KGameInfoFile = "gameinfo.dat";

static void extractField(const std::string& aString, std::string& aValue, int& aPos)
{
	int endPos = aPos;
	int maxLen = aString.size() - aPos;
	
	while ((maxLen > 0) && (aString[endPos] != '|'))
	{
		endPos++;
		maxLen--;
	}
	
	int len = endPos - aPos;
	if (len > 0)
		aValue.assign(aString, aPos, len);
	
	aPos = endPos +1;
}

static void ExtractCategoriesFromString(const std::string& aCategoriesAsString, std::vector<std::string>& aCategories)
{
	int startPos = 0;
	int length = aCategoriesAsString.size();
	while (startPos < length)
	{
		std::string category;
		int endPos = aCategoriesAsString.find(" / ", startPos);
		if (endPos != (int) std::string::npos)
		{
			category.assign(aCategoriesAsString, startPos, endPos - startPos);
			startPos += endPos + 3;
		}
		else
		{
			category.assign(aCategoriesAsString, startPos, length - startPos);
			startPos = length;
		}
		if (category.size() > 0)
			aCategories.push_back(category);
	}
}


std::string GameEngine::MapCategory(const std::string& aCategory, const TStringMap& aCategoryMapping)
{
	std::string category = (aCategory.size() == 0) ? "None" : aCategory;
	auto catIt = aCategoryMapping.find(category);
	if (catIt != aCategoryMapping.end())
		return(catIt->second);
	return category;
}

void GameEngine::LoadGamesData(const std::string& aName, const TStringMap& aCategoryMapping, TGameMap& aGameMap)
{
	FillGameMap(aName, aCategoryMapping, aGameMap);
	
	iGameInfoModel.Construct(iDataDir + Platform::PathSeperator() + aName, iUserDir, aGameMap, aName);
}

static void ReadGameInfoFile(const std::string& aGameInfoFile, TStringMap& aGameInfo)
{
	TStringMap gameInfo;
	FileUtils::ReadFile(aGameInfoFile, gameInfo);
	
	for (auto &game : gameInfo)
		aGameInfo[StringUtils::GetPureAlphaNumName(game.first)] = game.second;
}

void GameEngine::FillGameMap(const std::string& aName, const TStringMap& aCategoryMapping, TGameMap& aGameMap)
{
	std::string gameInfoFileName(iDataDir + Platform::PathSeperator() + aName + Platform::PathSeperator() + KGameInfoFile);
	
	TStringMap gameInfo;
	ReadGameInfoFile(gameInfoFileName, gameInfo);
	printf("%s: read game details for %d games\n", aName.c_str(), (int) gameInfo.size());

	if (iGameEngineProperties.iRomDirectory.iPathName.size() > 0)
	{
		iMediaDirectoryMapper.AddDirectory(iGameEngineProperties.iRomDirectory, iGameEngineProperties.iRomFileExtension);
		printf("%s: found %d roms in \"%s\"\n", aName.c_str(), (int) iMediaDirectoryMapper.GetMap(iGameEngineProperties.iRomDirectory.iPathName).size(), iGameEngineProperties.iRomDirectory.iPathName.c_str());
	}
	
	if (iGameEngineProperties.iIncludeAllRoms)
		FillGameMap_RomsAreLeading(aName, gameInfo);
	else
		FillGameMap_DatFileIsLeading(aName, gameInfo);

	CreateGameInfo(aName, aCategoryMapping, gameInfo, aGameMap);
}

void GameEngine::FillGameMap_DatFileIsLeading(const std::string& aName, TStringMap& aGameInfo)
{
	if (iGameEngineProperties.iRomDirectory.iPathName.size() > 0)
		RemoveUnsupportedGames(aGameInfo);
}

void GameEngine::FillGameMap_RomsAreLeading(const std::string& aName, TStringMap& aGameInfo)
{
	RemoveGamesWithoutRoms(aGameInfo);
	AddRomsNotInGamesDatFile(aGameInfo);
}


void GameEngine::GetGameInfo(const std::string& aGameName, std::string& aGameInfo)
{
	iGameInfoModel.GetGameInfo(aGameName, aGameInfo);
}

void GameEngine::CreateGameInfo(const std::string& aName, const TStringMap& aCategoryMapping, const TStringMap& aGameInfoList, TGameMap& aGameMap)
{
	for (auto &gameInfoEntry : aGameInfoList)
	{
		std::string field;
		TGame game;
		
		game.iEngine = aName;
		game.iName = gameInfoEntry.first;
		
		int pos = 0;
		extractField(gameInfoEntry.second, game.iDescription, pos);
		extractField(gameInfoEntry.second, game.iCloneOf, pos);
		extractField(gameInfoEntry.second, game.iYear, pos);
		extractField(gameInfoEntry.second, game.iManufacturer, pos);
		extractField(gameInfoEntry.second, field, pos);
		game.iPlayers = atoi(field.c_str());
		extractField(gameInfoEntry.second, field, pos); // ignore joyway
		extractField(gameInfoEntry.second, field, pos);
		game.iButtons = atoi(field.c_str());
		extractField(gameInfoEntry.second, field, pos);	//ignore coins
		extractField(gameInfoEntry.second, field, pos);
		game.iSimulatenousPlay = (field.find("sim") != std::string::npos);
		game.iAlternatePlay = (field.find("alt") != std::string::npos);
		
		std::string categories;
		std::vector<std::string> categoriesList;
		extractField(gameInfoEntry.second, categories, pos);
		ExtractCategoriesFromString(categories, categoriesList);
		if (categoriesList.size() > 0)
			game.iCategory = MapCategory(categoriesList.at(0), aCategoryMapping);
		else
			game.iCategory = MapCategory("", aCategoryMapping);
		
		aGameMap[game.iName] = game;
	}
	printf("%s: supported number of games=%d\n", aName.c_str(), (int) aGameMap.size());
}

void GameEngine::RemoveUnsupportedGames(TStringMap& aGameInfo)
{
	auto game=aGameInfo.begin();
	while(game!=aGameInfo.end())
	{
		if (!iMediaDirectoryMapper.HasMediaFile(iGameEngineProperties.iRomDirectory.iPathName, game->first))
			game = aGameInfo.erase(game);
		else
			++game;
	}
}

void GameEngine::RemoveGamesWithoutRoms(TStringMap& aGameInfo)
{
	int count = aGameInfo.size();
	auto game=aGameInfo.begin();
	while(game!=aGameInfo.end())
	{
		if (!iMediaDirectoryMapper.HasMediaFile(iGameEngineProperties.iRomDirectory.iPathName, game->first))
			game = aGameInfo.erase(game);
		else
			++game;
	}
	
	printf("Removed %d games without roms\n", count - (int) aGameInfo.size());
}


std::string GameEngine::GetDescription(const std::string& aFileName, char aIgnoreLeading, char aIgnoreTrailing)
{
	auto pos = aFileName.find_last_of(Platform::PathSeperator()[0]);
	if (pos != std::string::npos)
	{ 
		return PartialConvertName(aFileName.substr(pos+1), aIgnoreLeading, aIgnoreTrailing);
	}
	return PartialConvertName(aFileName, aIgnoreLeading, aIgnoreTrailing);
}

void GameEngine::AddRomsNotInGamesDatFile(TStringMap& aGameInfo)
{
//	for (auto &game : aGameInfo)
//		printf("GAMEINFO: game=%s\n", game.first.c_str());

	int count = 0;
	for (auto &entry : iMediaDirectoryMapper.GetMap(iGameEngineProperties.iRomDirectory.iPathName))
		if ((entry.first.size() > 0) && (aGameInfo.find(entry.first) == aGameInfo.end()))
		{
			std::string description = GetDescription(entry.second, iGameEngineProperties.iRomDirectory.GetIgnoreLeadingChar(), iGameEngineProperties.iRomDirectory.GetIngoreTrailingChar());
			aGameInfo[entry.first] = description + "||???|Unknown|1|||6||Unknown";
			count++;
		}
	
	printf("Added %d games with roms which were not in gameinfo.dat\n", (int) count);
}

std::string GameEngine::GetMediaFileName(const std::string& aGameName, const std::string& aDirectory)
{
	std::string result = iMediaDirectoryMapper.GetMediaFileName(aDirectory, aGameName);
//	printf("GameEngine::GetMediaFileName: game=%s, directory=%s, object=%s\n", aGameName.c_str(), aDirectory.c_str(), result.c_str());
	if (result.size() == 0)
		return aGameName;
	return aDirectory + Platform::PathSeperator() + result;
}

void GameEngine::PopulateMediaDirectoryMapper(const TMediaDirectory& aMediaDirectory, const TGameMap& aGameMap)
{
	iMediaDirectoryMapper.AddDirectory(aMediaDirectory, aGameMap, iUserDir);
}
