/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "gameinfomodel.h"
#include "platform.h"
#include "fileutils.h"
#include "stringutils.h"
#include <string.h>

const std::string KHistoryFile = "history.dat";
const std::string KHistoryIndexFile = "history.idx";

void GameInfoModel::Construct(const std::string& aDataDirectory, const std::string& aUserDirectory, const TGameMap& aGameMap, const std::string& aGameEngine)
{
	iInfoFileName.assign(aDataDirectory + Platform::PathSeperator() + KHistoryFile);
	iInfoIndexFileName.assign(aUserDirectory + Platform::PathSeperator() + aGameEngine + "_" + KHistoryIndexFile);
	
	if (!FileUtils::FileExists(iInfoIndexFileName))
		BuildIndexFile(aGameMap);
	ReadIndexFile();
}

template<>
void Store<GameInfoModel::TGameInfoIndex>(GameInfoModel::TGameInfoIndex& aContainer, std::string& aKey, std::string& aValue)
{
	std::istringstream is_line(aValue);
	std::string number1;
	std::string number2;
	std::getline(is_line, number1, ',');
	std::getline(is_line, number2, ',');

	
	GameInfoModel::TGameInfoIndexEntry gameInfoIndexEntry = {std::atoi(number1.c_str()), std::atoi(number2.c_str())};
	aContainer[aKey] = gameInfoIndexEntry;
}

void GameInfoModel::ReadIndexFile()
{
	FileUtils::ReadFileNew(iInfoIndexFileName, iGameInfoIndex);
}

void GameInfoModel::WriteGameInfoIndexEntry(std::ofstream& aFileStream, const std::vector<std::string>& aGameNames, const TGameInfoIndexEntry& aGameInfoIndexFile)
{
	for (auto &gameName : aGameNames)
		aFileStream << gameName << "=" << aGameInfoIndexFile.iOffset << "," << aGameInfoIndexFile.iLength << std::endl;
}


void GameInfoModel::ExtractGameNames(std::string& aValue, std::vector<std::string>& aGameNames)
{
	aGameNames.clear();
	StringUtils::StripWhiteSpaces(aValue);
	std::istringstream is_line(aValue);
	std::string name;
	while (std::getline(is_line, name, ','))
		aGameNames.push_back(name);
}

void GameInfoModel::CreateSupportedGames(const std::vector<std::string>& aGameNames, std::vector<std::string>& aSupportedGameNames, const TGameMap& aGameMap)
{
	aSupportedGameNames.clear();
	for (auto &gameName : aGameNames)
		if (aGameMap.find(gameName) != aGameMap.end())
			aSupportedGameNames.push_back(gameName);
}

void GameInfoModel::BuildIndexFile(const TGameMap& aGameMap)
{
	std::ifstream fileStream(iInfoFileName);
	if (fileStream)
	{
		std::ofstream indexFileStream(iInfoIndexFileName);

		std::string line;
		std::vector<std::string> gameNames;
		TGameInfoIndexEntry gameInfoIndexEntry = { 0, 0 };
		while (std::getline(fileStream, line))
		{
			if (line.length() > 0)
			{
				if (line.compare(0, 4, "$end") == 0)
				{
					gameInfoIndexEntry.iLength = int(fileStream.tellg()) - gameInfoIndexEntry.iOffset;
					WriteGameInfoIndexEntry(indexFileStream, gameNames, gameInfoIndexEntry);
					gameNames.clear();
				}
				else if (line.compare(0, 5, "$info") == 0)
				{
					std::istringstream is_line(line);
					std::string key;
					
					std::getline(is_line, key, '=');
					if (key.size() > 0)
					{
						std::string value;
						std::getline(is_line, value);
						if (value.size() > 0)
						{
							std::vector<std::string> rawGameNames;
							ExtractGameNames(value, rawGameNames);
							CreateSupportedGames(rawGameNames, gameNames, aGameMap);
							gameInfoIndexEntry.iOffset = (int) fileStream.tellg();
						}
					}
				}
			}
		}
	}
}


void GameInfoModel::GetGameInfo(const std::string& aGameName, std::string& aGameInfo)
{
	aGameInfo.clear();
	
	auto it = iGameInfoIndex.find(aGameName);
	if (it != iGameInfoIndex.end())
		FileUtils::ReadFilePart(iInfoFileName, it->second.iOffset, it->second.iLength, aGameInfo);
}
