/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "gameenginemanager.h"
#include "gameengine.h"

#include "platform.h"
#include "fileutils.h"
#include "systemproperties.h"

GameEngineManager::~GameEngineManager()
{
	for (auto &it : iGameEngines)
		delete it.second;
	iGameEngines.clear();
}

void GameEngineManager::Construct()
{
	for (auto &engine : AppSystemProperties.iGameEnginesProperties)
		AddGameEngine(engine.first, new GameEngine());
}

void GameEngineManager::AddGameEngine(const std::string& aName, GameEngine* aGameEngine)
{
	iGameEngines[aName] = aGameEngine;
	iGamesContainer.emplace(aName, GamesContainer(aName, iGameFilterConfig));
	
	printf("********** Game engine: %s **********\n", aName.c_str());
	aGameEngine->Construct(AppSystemProperties.iGameEnginesProperties[aName], AppSystemProperties.iDataDir, AppSystemProperties.iUserDir);
	
	TStringMap	categoryMapping;
	LoadCategoryMap(aName, categoryMapping);
	aGameEngine->LoadGamesData(aName, categoryMapping, iGamesContainer.at(aName).iGameMap);

	if (AppSystemProperties.iGameEnginesProperties[aName].iMediaDirectory.size() > 0)
	{
		for (int i=0 ; i<KNumOfBuiltinMediaDirectories; i++)
		{
			AppSystemProperties.iGameEnginesProperties[aName].iMediaDirectories[KNumOfUserMediaDirectories + i].iName = std::string("dir_") + KBuiltinDirectoryNames[i];
			AppSystemProperties.iGameEnginesProperties[aName].iMediaDirectories[KNumOfUserMediaDirectories + i].iPathName = 
				AppSystemProperties.iGameEnginesProperties[aName].iMediaDirectory + Platform::PathSeperator() + KBuiltinDirectoryNames[i];
		}
	}
	
	for (int i=0 ; i<KTotalNumOfMediaDirectories ; i++)
	{
		if (AppSystemProperties.iGameEnginesProperties[aName].iMediaDirectories[i].iPathName.size() > 0)
			aGameEngine->PopulateMediaDirectoryMapper(AppSystemProperties.iGameEnginesProperties[aName].iMediaDirectories[i], iGamesContainer.at(aName).iGameMap);
	}
	
	printf("**************************************\n");
}

GameEngine& GameEngineManager::GetGameEngine(const std::string& aName) const
{
	static GameEngine emptyGameEngine;
	auto it = iGameEngines.find(aName);
	if ( it != iGameEngines.end())
		return *(it->second);
	return emptyGameEngine;
}


void GameEngineManager::LoadCategoryMap(const std::string& aEngineName, TStringMap& aCategoryMapping)
{
	std::string catMapFileName(std::string("data") + Platform::PathSeperator() + aEngineName + Platform::PathSeperator() + "catmap.ini");
	
	FileUtils::ReadFile(catMapFileName, aCategoryMapping);
}

void GameEngineManager::GetGameInfo(const std::string& aGameID, std::string& aGameInfo)
{
	std::string gameName;
	std::string engineName;
	TGame::DecodeGameID(aGameID, gameName, engineName);
	
	GetGameEngine(engineName).GetGameInfo(gameName, aGameInfo);
}


const TGame* GameEngineManager::GetGame(const std::string& aGameID)
{
	std::string gameName;
	std::string engineName;
	TGame::DecodeGameID(aGameID, gameName, engineName);
	
	auto itContainer = iGamesContainer.find(engineName);
	if (itContainer != iGamesContainer.end())
	{
		auto itGame = itContainer->second.find(gameName);
		if (itGame != itContainer->second.end())
			return &(*itGame);
	}
	return nullptr;
}

std::string GameEngineManager::GetMediaFileName(const std::string& aGameName, const std::string& aEngineName, const std::string& aLogicalDirectory)
{
	const std::string& path = AppSystemProperties.iGameEnginesProperties[aEngineName].GetDirectoryPath(aLogicalDirectory);
	return GetGameEngine(aEngineName).GetMediaFileName(aGameName, path);
}

std::string GameEngineManager::GetMediaFileName(const TGame& aGame, const std::string& aLogicalDirectory)
{
	return GetMediaFileName(aGame.iName, aGame.iEngine, aLogicalDirectory);
}

std::string GameEngineManager::GetMediaFileName(const std::string& aGameId, const std::string& aLogicalDirectory)
{
	std::string gameName;
	std::string engineName;
	
	TGame::DecodeGameID(aGameId, gameName, engineName);
	
	return GetMediaFileName(gameName, engineName, aLogicalDirectory);
}


std::string GameEngineManager::GetRomFileName(const std::string& aGameId)
{
	std::string gameName;
	std::string engineName;
	
	TGame::DecodeGameID(aGameId, gameName, engineName);

	return GetGameEngine(engineName).GetMediaFileName(gameName, AppSystemProperties.iGameEnginesProperties[engineName].iRomDirectory.iPathName);
}
