/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "gamerunner.h"
#include "fileutils.h"
#include "stringutils.h"
#include "systemproperties.h"
#include "userproperties.h"
#include "appmenudefs.h"
#include "sdlmenu2dmodel.h"
#include "platform.h"
#include "sdleventmanager.h"
#include "settingsdialogs.h"
#include "controllerutils.h"
#include "sdlinfoprint.h"
#include "gameenginemanager.h"


namespace GameRunner
{

	std::string Model::KFileName("recent.rc");

	std::string Model::CreateCommandLine(const std::string& aGameName, const std::string& aEngineName)
	{
		std::string commandLine = AppSystemProperties.iGameEnginesProperties[aEngineName].iCommandLine;

		StringUtils::TSubstituteMap substituteMap;

		substituteMap["%game%"] = aGameName;
		substituteMap["%romfile%"] = iGameEngineManager.GetRomFileName(iGameIDToRun);
		StringUtils::Substitute(commandLine, substituteMap);

		return commandLine;
	}


	void Model::Construct()
	{
		TStringPairList stringPairList;
		FileUtils::ReadFile(AppSystemProperties.iUserDir + Platform::PathSeperator() + KFileName, stringPairList);
		iMenu = sdl::Menu2DModel::ConstructMenu(KMenuNameRecentGames, KMenuTypeRecent, stringPairList);
		Reconfigure();
	}

	void Model::Reconfigure()
	{
		if (iMenu)
		{
			iMenu->SetEnabled(AppUserProperties.iRecentGames.iEnabled);
			iMenu->PruneMenuItems(AppUserProperties.iRecentGames.iMax);
		}
	}


	bool Model::WantRunGame()
	{
		return !iGameIDToRun.empty();
	}

	void Model::SetRunGame(const sdl::MenuItem& aMenuItem)
	{
		iGameIDToRun = aMenuItem.GetCommand();
		AddToRecent(aMenuItem);
		sdl::EventManager::SendCommand(sdl::TCommand::EQuit);
	}

	bool Model::RunGame()
	{
		std::string gameName;
		std::string engineName;

		TGame::DecodeGameID(iGameIDToRun, gameName, engineName);

		std::string commandLine = CreateCommandLine(gameName, engineName);
		printf("Model::RunGame=%s %s\n", AppSystemProperties.iGameEnginesProperties[engineName].iExeName.c_str(), commandLine.c_str());

		std::vector<std::string> stringList;
		StringUtils::ParseCommandLine(commandLine, stringList);

		bool success = Platform::RunProgram(AppSystemProperties.iGameEnginesProperties[engineName].iExeName, stringList);

		iGameIDToRun.clear();

		return success;
	}

	void Model::AddToRecent(const sdl::MenuItem& aMenuItem)
	{
		if (iMenu)
		{
			iMenu->DeleteMenuItem(aMenuItem.GetCommand(), sdl::Menu::TNotifiyPolicy::EDontNotifyClients); // only one entry at the beginning of the list
			iMenu->AddMenuItem(
				new sdl::MenuItem(aMenuItem.GetName().c_str(), aMenuItem.GetCommand().c_str()),
				sdl::Menu::TAddPolicy::EAddToFront,
				sdl::Menu::TNotifiyPolicy::EDontNotifyClients
				);
			iMenu->PruneMenuItems(AppUserProperties.iRecentGames.iMax, sdl::Menu::TNotifiyPolicy::EDontNotifyClients);
			Persist();
		}
	}

	void Model::Reset()
	{
		if (iMenu)
		{
			iMenu->DeleteAllMenuItems();
			Persist();
		}
	}

	void Model::Persist()
	{
		if (iMenu)
		{
			TStringPairList stringPairList;
			sdl::Menu2DModel::GetMenuList(*iMenu, stringPairList);
			FileUtils::WriteFile(AppSystemProperties.iUserDir + Platform::PathSeperator() + KFileName, stringPairList);
		}
	}


	void RecentGamesMenuController::MenuItemCommand(const sdl::MenuItem& aMenuItem, const sdl::TCommand& aCommand)
	{
		if ((aCommand == sdl::ESelect) && !aMenuItem.GetCommand().empty())
			iModel.SetRunGame(aMenuItem);
	}

	void SettingMenuController::MenuItemCommand(const sdl::MenuItem& aMenuItem, const sdl::TCommand& aCommand)
	{
		if (aCommand == sdl::ESelect)
		{
			if (aMenuItem.GetName().compare(KMenuCmdClearRecentGames) == 0)
				CmdClearRecentGames();
			else if (aMenuItem.GetName().compare(KMenuCmdConfigureRecentGamesMenu) == 0)
				CmdConfigureRecentGamesMenu();
		}
	}

	void SettingMenuController::CmdClearRecentGames()
	{
		if (sdl::StdDialogs::MessageDialog("Clearing recent games", "Are you sure?"))
		{
			iModel.Reset();
			sdl::InfoPrint::Message("Recent games cleared");
		}
	}

	void SettingMenuController::CmdConfigureRecentGamesMenu()
	{
		DialogConfigureSystemMenu dialogSettingsRecentGamesMenu(AppUserProperties.iRecentGames.iEnabled, AppUserProperties.iRecentGames.iMax, std::string("recent games"));
		dialogSettingsRecentGamesMenu.SetTitle(KMenuCmdConfigureRecentGamesMenu);
		if (dialogSettingsRecentGamesMenu.Execute())
		{
			ControllerUtils::SaveResourceFileUserProperties();
			iModel.Reconfigure();
		}
	}
}