/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlrenderer.h"
#include "commonsdlrenderer.h"

namespace sdl
{
	class SDLTexture : public ITexture
	{
	public:
		SDLTexture(SDL_Texture* aTexture = nullptr) : iTexture(aTexture) {}
		SDLTexture(SDL_Renderer& aRenderer, const TSize& aSize, int aTextureFormat, bool aRenderingTexture = false);
		virtual ~SDLTexture() override;
		virtual TRect GetRect() override;
		virtual void SetColor(const TColor& aColor) override;
		virtual void CopyBytes(void* aBytes, int aBytesPerRow) override;
		
		SDL_Texture* iTexture = nullptr;
	};
	
	class SDLRenderer : public IRenderer
	{
	public:
		SDLRenderer(const TSize& aSize, const std::string& aWindowTitle, bool aFullScreen);
		virtual ~SDLRenderer() override;
		
		virtual ITexture* 	CreateTextureFromSVG(const std::string& aSVGText) override;
		virtual ITexture* 	CreateTextureFromTTF(TTF_Font& aFont, const std::string& aText, const TColor& aColor) override;
		virtual ITexture* 	CreateTextureFromFile(const std::string& aFileName) override;
		
		virtual ITexture*	CreateMemoryTexture(const TSize& aSize, const char* aChromaFormat = nullptr) override;
		virtual ITexture*	CreateRenderingTexture(const TSize& aSize) override;
		
		virtual IRenderer*	CreateRenderer(ITexture& aTexture) override;
		
		virtual void RenderTexture(ITexture& aTexture, const TPoint& aPoint) override;
		virtual void RenderTextureRotated(ITexture& aTexture, const TPoint& aPoint, float aDegrees=0) override;
		virtual void RenderTextureScaled(ITexture& aTexture, const TPoint& aPoint, float aScaleX, float aScaleY) override;
		virtual void RenderFillRect(const TRect& aRect, const TColor& aColor) override;
		virtual void RenderRect(const TRect& aRect, const TColor& aColor) override;
		
		virtual void Clear() override;
		
	protected:
		SDLRenderer() {}
		SDL_Renderer* 	iRenderer = nullptr;
		
		SDL_Texture* CreateTextureFromSurface(SDL_Surface* aSurface);
		
		virtual TRect SetClipRect(const TRect& aRect) override;
		
		virtual void Present() override;
		
	private:
		SDL_Window*		iWindow = nullptr;
	};
}


using namespace sdl;


static void DumpRendererInfo(SDL_Renderer& aRenderer)
{
	SDL_RendererInfo rendererInfo;
	memset(&rendererInfo, 0, sizeof(rendererInfo));
	SDL_GetRendererInfo(&aRenderer, &rendererInfo);
	printf("SDL rendering info:\n");
	printf("  renderer=%s\n", rendererInfo.name);
	if (rendererInfo.flags != 0)
	{
		std::string flagsText;
		if (rendererInfo.flags & SDL_RENDERER_SOFTWARE)
			flagsText = "software";
		if (rendererInfo.flags & SDL_RENDERER_ACCELERATED)
			flagsText += (flagsText.size() == 0) ? "accelerated" : ", accelerated";
		if (rendererInfo.flags & SDL_RENDERER_PRESENTVSYNC)
			flagsText += (flagsText.size() == 0) ? "present v-sync" : ", present v-sync";
		if (rendererInfo.flags & SDL_RENDERER_TARGETTEXTURE)
			flagsText += (flagsText.size() == 0) ? "render to texture" : ", render to texture";
		printf("  flags=%s\n", flagsText.c_str());
	}
}

/******************************************************
 * 
 * SDLTextureRenderer
 * 
 *****************************************************/

class SDLTextureRenderer : public SDLRenderer
{
public:
	SDLTextureRenderer(SDL_Renderer* aRenderer, SDLTexture& aTexture)
	{
		iOldTexture = SDL_GetRenderTarget(aRenderer);
		
		if (SDL_SetRenderTarget(aRenderer, aTexture.iTexture) == 0)
		{
			iRenderer = aRenderer;
		}
		else
		{
			printf("SDLTextureRenderer::SDLTextureRenderer SDL_SetRenderTarget failed: %s\n", SDL_GetError());
		}
	}
	
	~SDLTextureRenderer()
	{
		SDL_SetRenderTarget(iRenderer, iOldTexture);
		iRenderer = nullptr; // important, otherwise SDLRenderer::~SDLRenderer will delete iRenderer
	}
	
	SDL_Texture* iOldTexture = nullptr;
};

/******************************************************
 * 
 * SDLTexture
 * 
 *****************************************************/
SDLTexture::SDLTexture(SDL_Renderer& aRenderer, const TSize& aSize, int aTextureFormat, bool aRenderingTexture)
{
	int textureAccess = aRenderingTexture ? SDL_TEXTUREACCESS_TARGET : SDL_TEXTUREACCESS_STREAMING;
	iTexture = SDL_CreateTexture(&aRenderer, aTextureFormat, textureAccess, aSize.iWidth, aSize.iHeight);
}


SDLTexture::~SDLTexture()
{
	if (iTexture)
	{
		SDL_DestroyTexture(iTexture);
		iTexture = nullptr;
	}
}


void SDLTexture::CopyBytes(void* aBytes, int aBytesPerRow)
{
	int pitch;
	void* rawTextureMemory;
	
	TRect rect = GetRect();
	if (SDL_LockTexture(iTexture, NULL, &rawTextureMemory, &pitch) != 0)
		printf("SDLTexture::CopyBytes: SDL_LockTexture failed: %s\n", SDL_GetError());
	
	memcpy(rawTextureMemory, aBytes, aBytesPerRow * rect.Height());
	
	SDL_UnlockTexture(iTexture);
}

TRect SDLTexture::GetRect()
{
	int width = 0;
	int height = 0;
	
	TRect rect(TPoint(0, 0));
	if (SDL_QueryTexture(iTexture, nullptr, nullptr, &width, &height) != 0)
		printf("SDLTexture::GetRect: SDL_QueryTexture failed, error=%s\n", SDL_GetError());
	
	return TRect(TSize(width, height));
}


void SDLTexture::SetColor(const TColor& aColor)
{
	SDL_SetTextureColorMod(iTexture, aColor.GetR(), aColor.GetG(), aColor.GetB());
	SDL_SetTextureAlphaMod(iTexture, aColor.GetA());
}


/******************************************************
 * 
 * SDLRenderer
 * 
 *****************************************************/


SDLRenderer::SDLRenderer(const TSize& aSize, const std::string& aWindowTitle, bool aFullScreen)
{
	Uint32 sdlWindowFlags = SDL_WINDOW_SHOWN;
	if (aFullScreen)
		sdlWindowFlags = SDL_WINDOW_FULLSCREEN_DESKTOP;
	
	iWindow = SDL_CreateWindow(aWindowTitle.c_str(),
								SDL_WINDOWPOS_CENTERED,
						SDL_WINDOWPOS_CENTERED,
						aSize.iWidth, aSize.iHeight,
						sdlWindowFlags);
	if(iWindow == nullptr)
	{
		printf("SDL_CreateWindow failed: %s\n", SDL_GetError());
		exit(-1);
	}
	
	iRenderer = SDL_CreateRenderer(iWindow, -1, 0);
	if (iRenderer  == nullptr)
	{
		printf("SDL_CreateRenderer failed: %s\n", SDL_GetError());
		exit(-1);
	}
	
	DumpRendererInfo(*iRenderer);
	
	if (!SDL_SetHint( SDL_HINT_RENDER_SCALE_QUALITY, "1"))
	{
		printf("Warning: linear texture filtering not enabled!\n"); 
	}
}

SDLRenderer::~SDLRenderer()
{
	if (iRenderer != nullptr)
	{
		SDL_DestroyRenderer(iRenderer);
		iRenderer= nullptr;
	}
	
	if (iWindow != nullptr)
	{
		SDL_DestroyWindow(iWindow);
		iWindow = nullptr;
	}
}


IRenderer* SDLRenderer::CreateRenderer(ITexture& aTexture)
{
	SDLTexture& texture = (SDLTexture&) aTexture;
	return new SDLTextureRenderer(iRenderer, texture);
}

ITexture* SDLRenderer::CreateTextureFromSVG(const std::string& aSVGText)
{
	void* surfaceData = nullptr;
	SDL_Surface* surface = CreateSurfaceFromSVG(aSVGText, &surfaceData);
	ITexture* texture = new SDLTexture(CreateTextureFromSurface(surface));
	free(surfaceData);
	return texture;
}

ITexture* SDLRenderer::CreateTextureFromTTF(TTF_Font& aFont, const std::string& aText, const TColor& aColor)
{
	SDL_Surface* surface = CreateSurfaceFromTTF(aFont, aText, aColor);
	return new SDLTexture(CreateTextureFromSurface(surface));
}

ITexture* SDLRenderer::CreateTextureFromFile(const std::string& aFileName)
{
	SDL_Surface* surface = CreateSurfaceFromFile(aFileName);
	return new SDLTexture(CreateTextureFromSurface(surface));
}

static std::map<std::string, int> sChromaFormatMapping = {{"RGBA", SDL_PIXELFORMAT_BGR888}};

static int GetTextureFormat(const char* aChromaFormat)
{
	int format = SDL_PIXELFORMAT_RGBA8888;
	if (aChromaFormat)
	{
		auto it = sChromaFormatMapping.find(aChromaFormat);
		if (it != sChromaFormatMapping.end())
			format = it->second;
	}
	return format;
}

ITexture* SDLRenderer::CreateMemoryTexture(const TSize& aSize, const char* aChromaFormat)
{
	return new SDLTexture(*iRenderer, aSize, GetTextureFormat(aChromaFormat));
}

ITexture* SDLRenderer::CreateRenderingTexture(const TSize& aSize)
{
	return new SDLTexture(*iRenderer, aSize, SDL_PIXELFORMAT_RGBA8888, true);
}

void SDLRenderer::RenderTexture(ITexture& aTexture, const TPoint& aPoint)
{
	SDLTexture& sdlTexture = (SDLTexture&) aTexture;
	TRect rect = aTexture.GetRect() + aPoint;
	SDL_Rect sdlRect = rect;
	if (SDL_RenderCopy(iRenderer, sdlTexture.iTexture, nullptr, &sdlRect) != 0)
		printf("SDLRenderer::DrawTexture: SDL_RenderCopy failed, error=%s\n", SDL_GetError());
}


void SDLRenderer::RenderTextureRotated(ITexture& aTexture, const TPoint& aPoint, float aDegrees)
{
	SDLTexture& sdlTexture = (SDLTexture&) aTexture;
	TRect rect = aTexture.GetRect() + aPoint;
	SDL_Rect sdlRect = rect;
	if (SDL_RenderCopyEx(iRenderer, sdlTexture.iTexture, nullptr, &sdlRect, aDegrees, NULL, SDL_FLIP_NONE) != 0)
		printf("SDLRenderer::RenderTextureRotated: SDL_RenderCopyEx failed, error=%s\n", SDL_GetError());
}

void SDLRenderer::RenderTextureScaled(ITexture& aTexture, const TPoint& aPoint, float aScaleX, float aScaleY)
{
	SDLTexture& sdlTexture = (SDLTexture&) aTexture;
	TRect rect = aTexture.GetRect() + aPoint;
	SDL_Rect sdlRect = rect;
	sdlRect.w *= aScaleX;
	sdlRect.h *= aScaleY;
	if (SDL_RenderCopy(iRenderer, sdlTexture.iTexture, nullptr, &sdlRect) != 0)
		printf("SDLRenderer::RenderTextureScaled: SDL_RenderCopy failed, error=%s\n", SDL_GetError());
}

void SDLRenderer::RenderFillRect(const TRect& aRect, const TColor& aColor)
{
	SDL_Rect sdlRect = aRect;
	SDL_SetRenderDrawColor(iRenderer, aColor.GetR(), aColor.GetG(), aColor.GetB(), 0xFF);
	SDL_RenderFillRect(iRenderer, &sdlRect);
}

void SDLRenderer::RenderRect(const TRect& aRect, const TColor& aColor)
{
	SDL_SetRenderDrawColor(iRenderer, aColor.GetR(), aColor.GetG(), aColor.GetB(), 0xFF);
	SDL_Rect sdlRect = aRect;
	SDL_RenderDrawRect(iRenderer, &sdlRect);
}

void SDLRenderer::Clear()
{
	SDL_SetRenderDrawColor(iRenderer, 0, 0, 0, 0xFF);
	SDL_RenderClear(iRenderer);
}

SDL_Texture* SDLRenderer::CreateTextureFromSurface(SDL_Surface* aSurface)
{
	SDL_Texture* texture = nullptr;
	if (aSurface)
	{
		texture = SDL_CreateTextureFromSurface(iRenderer, aSurface);
		if (texture == nullptr)
			printf("SDLRenderer::CreateTextureFromSurface SDL_CreateTextureFromSurface failed\n");
		
		SDL_FreeSurface(aSurface);
	}
	
	return texture;
}

TRect SDLRenderer::SetClipRect(const TRect& aRect)
{
	SDL_Rect oldClipRect;
	SDL_RenderGetClipRect(iRenderer, &oldClipRect);

	if (aRect.IsEmpty())
	{
		if (SDL_RenderSetClipRect(iRenderer, nullptr) < 0)
			printf("SDLRenderer::SetClipRect: SDL_RenderSetClipRect (nullptr) failed, error=%s\n", SDL_GetError());
	}
	else
	{
		SDL_Rect rect = aRect;
		if (SDL_RenderSetClipRect(iRenderer, &rect) < 0)
			printf("SDLRenderer::SetClipRect: SDL_RenderSetClipRect failed, error=%s\n", SDL_GetError());
	}

	return TRect(oldClipRect);
}

void SDLRenderer::Present()
{
	SDL_RenderPresent(iRenderer); 
}


IRenderer* sdl::CreateSDL2Renderer(const TSize& aSize, const std::string& aWindowTitle, bool aFullScreen)
{
	return new SDLRenderer(aSize, aWindowTitle, aFullScreen);
}