/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlaudio.h"
#include "sdlcore.h"
#include "sdluserproperties.h"
#include "platform.h"
#include "fileutils.h"
#include "algorithm"

using namespace sdl;

#define AUDIO_CHANNEL 1

static std::map<Audio::TAudioTypes, Audio::TAudioData > sAudioData = 
{
	{Audio::TAudioTypes::EIntro,		{"Startup", "intro"}},
	{Audio::TAudioTypes::EOk,			{"Ok", "ok"}},
	{Audio::TAudioTypes::ECancel,		{"Cancel", "cancel"}},
	{Audio::TAudioTypes::EScrollH,		{"Horizontal scroll", "scrollh"}},
	{Audio::TAudioTypes::EScrollV,		{"Vertical scroll", "scrollv"}},
	{Audio::TAudioTypes::EScrollEnd,	{"End of list", "scrollend"}},
};

static std::map<int, Audio::TAudioTypes> sAudioEventMap = 
{
	{(int) UserFeedbackManager::TEvent::EUserInputOk,		Audio::TAudioTypes::EOk},
	{(int) UserFeedbackManager::TEvent::EUserInputCancel,	Audio::TAudioTypes::ECancel},
	{(int) UserFeedbackManager::TEvent::EDialogOk,			Audio::TAudioTypes::EOk},
	{(int) UserFeedbackManager::TEvent::EDialogCancel,		Audio::TAudioTypes::ECancel},
	{(int) UserFeedbackManager::EUserStartup,				Audio::TAudioTypes::EIntro},
	{(int) UserFeedbackManager::EMenuVerticalScroll, 		Audio::TAudioTypes::EScrollV},
	{(int) UserFeedbackManager::EMenuHorizontalScroll, 		Audio::TAudioTypes::EScrollH},
	{(int) UserFeedbackManager::EMenuEndList, 				Audio::TAudioTypes::EScrollEnd},
};


void Audio ::Construct()
{
	if(Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, 2, SDLUserProperties.iAudioProperties.iSamples) == 0) 
	{
		LoadChunks();
		Mix_AllocateChannels(16);
		SetVolume(SDLUserProperties.iAudioProperties.iVolume);
	}
	else
	{
		printf("Warning: Mix_OpenAudio failed (%s) - there will be no audio\n", Mix_GetError());
	}
}

void Audio::SetVolume(int aPercentage)
{
	if (aPercentage < 0)
		aPercentage = 0;
	if (aPercentage > 100)
		aPercentage = 100;
	
	int volume = (MIX_MAX_VOLUME *  aPercentage) / 100;
	Mix_Volume(AUDIO_CHANNEL, volume);
}


void Audio::GetAudioData(std::vector<Audio::TAudioData>& aAudioData)
{
	aAudioData.clear();
	for (auto &audioData : sAudioData)
		aAudioData.push_back(audioData.second);
}

std::string Audio::GetAudioFileName(TAudioTypes& aAudioType)
{
	std::string fileName;
	FileUtils::FindFileInDirs(sAudioData[aAudioType].iFileName + ".wav", System::AudioDirs.GetDirectoryNames(), fileName);
	return fileName;
}


Mix_Chunk* Audio::LoadChunk(TAudioTypes& aAudioType)
{
	std::string filename = GetAudioFileName(aAudioType);
	Mix_Chunk* chunk = Mix_LoadWAV(filename.c_str());
	if (chunk == nullptr)
		printf("Warning: Mix_LoadWAV %s failed (%s)\n", filename.c_str(), Mix_GetError());
	return chunk;
}

void Audio::LoadChunks()
{
	for (int i=0; i< (int)TAudioTypes::ELast ; i++)
	{
		Audio::TAudioTypes audioType = (Audio::TAudioTypes) i;
		iChunkMap[audioType] = LoadChunk(audioType);
	}
}

void Audio::ReleaseChunks()
{
	std::for_each(iChunkMap.begin(), iChunkMap.end(), [=](std::pair<TAudioTypes, Mix_Chunk*> aEntry) {if (aEntry.second) Mix_FreeChunk(aEntry.second); });
}


bool Audio::IsPlaying()
{
	return (Mix_Playing(AUDIO_CHANNEL) != 0);
}

bool Audio::IsUserConfigured(TAudioTypes aAudioType)
{
	bool isUserConfigured = false;
	std::string fileName = sAudioData[aAudioType].iFileName;
	for (auto &event : SDLUserProperties.iAudioProperties.iEnabledEvents)
	{
		if ((event.compare("all") == 0) || (event.compare(fileName) == 0))
		{
			isUserConfigured = true;
			break;
		}
	}
	return isUserConfigured;
}

bool Audio::ShouldPlay(TAudioTypes aAudioType)
{
	bool shouldPlay = !SDLUserProperties.iAudioProperties.iMute && IsUserConfigured(aAudioType);

	if (shouldPlay)
	{
		if (IsPlaying() && (iCurrentAudioType == TAudioTypes::EIntro))
			shouldPlay = false;
		else if (IsPlaying())
			shouldPlay = iCurrentAudioType != aAudioType;
	}
	
	return shouldPlay;
}

void Audio::Play(TAudioTypes aAudioType)
{
	if (ShouldPlay(aAudioType))
	{
		Mix_Chunk* chunk = iChunkMap[aAudioType];
		if (chunk != nullptr)
		{
			iCurrentAudioType = aAudioType;
			if( Mix_PlayChannel(AUDIO_CHANNEL, chunk, 0) == -1) 
				printf("Warning, Mix_PlayChannel failed: %s\n", Mix_GetError());
		}
	}
}

void Audio::HandleEvent(const UserFeedbackManager::TEvent& aEvent)
{
	TAudioTypes audioType = TAudioTypes::ELast;
	auto it = sAudioEventMap.find((int) aEvent);
	if (it != sAudioEventMap.end())
		audioType = it->second;
	if (int(audioType) >= 0 && int(audioType)<(int)TAudioTypes::ELast)
		Play(audioType);
}
