/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#ifndef __SDL_BASE_H
#define __SDL_BASE_H

#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>
#include <SDL2/SDL_ttf.h>
#include <SDL2/SDL2_gfxPrimitives.h>
#include "SDL_gpu.h"

#include <map>
#include <string>
#include <vector>
#include <memory>
#include <algorithm>

namespace sdl
{
	
inline void OverrideNonEmptyString(std::string& aString, const std::string& aOther)
{
	if (aOther.size() > 0)
		aString = aOther;
}

inline void OverridePositiveValue(int& aValue, const int aOther)
{
	if (aOther > 0)
		aValue = aOther;
}

inline void SetDefaultString(std::string& aString, const std::string& aOther)
{
	if (aString.size() == (int) 0)
		aString = aOther;
}

inline void SetDefaultPositiveValue(int& aValue, const int aOther)
{
	if (aValue <= 0)
		aValue = aOther;
}


class TMargin
{
public:
	TMargin() : TMargin(0) { }
	TMargin(int aMargin) { iLeft = iRight = iTop = iBottom = aMargin; }
	TMargin(int aLeft, int aTop, int aRight, int aBottom) { iLeft = aLeft; iRight = aRight; iTop = aTop; iBottom = aBottom; }
	int iLeft;
	int iTop;
	int iRight;
	int iBottom;
	
	void OverrideDefaults(const TMargin& aOther)
	{
		OverridePositiveValue(iLeft, aOther.iLeft);
		OverridePositiveValue(iTop, aOther.iTop);
		OverridePositiveValue(iRight, aOther.iRight);
		OverridePositiveValue(iBottom, aOther.iBottom);
	}
};


class TPoint
{
public:
	TPoint(float aXY = 0) : iX(aXY), iY(aXY) {}
	TPoint(float aX, float aY) : iX(aX), iY(aY) {}
	TPoint(const SDL_Point& aSdlPoint) : iX(aSdlPoint.x), iY(aSdlPoint.y) {}
	
	operator SDL_Point() const { SDL_Point point = {(int) iX, (int) iY}; return point; }
	
	TPoint& operator+=(const TPoint& aOther) { iX += aOther.iX; iY += aOther.iY; return *this; }
	friend TPoint operator+(TPoint aPoint, const TPoint& aOther) { aPoint+=aOther; return aPoint; }
	TPoint& operator-=(const TPoint& aOther) { iX -= aOther.iX; iY -= aOther.iY; return *this; }
	friend TPoint operator-(TPoint aPoint, const TPoint& aOther) { aPoint-=aOther; return aPoint; }

	bool operator==(const TPoint& aOther) const { return (iX == aOther.iX) && (iY == aOther.iY); }
	
	float iX = 0;
	float iY = 0;
};

class TSize
{
public:
	TSize(float aSize = 0) : iWidth(aSize), iHeight(aSize) { }
	TSize(float aWidth, float aHeight) : iWidth(aWidth), iHeight(aHeight) { }
	
	TSize& operator+=(const TSize& aOther) { iWidth += aOther.iWidth; iHeight += aOther.iHeight; return *this; }
	friend TSize operator+(TSize aSize, const TSize& aOther) { aSize+=aOther; return aSize; }
	TSize& operator-=(const TSize& aOther) { iWidth -= aOther.iWidth; iHeight -= aOther.iHeight; return *this; }
	friend TSize operator-(TSize aSize, const TSize& aOther) { aSize-=aOther; return aSize; }
	
	bool operator==(const TSize& aOther) const { return (iWidth == aOther.iWidth) && (iHeight == aOther.iHeight); }

	bool IsEmpty() const {return iWidth <= 0 || iHeight <= 0; }
	
	float iWidth = 0;
	float iHeight = 0;
};


class TRect
{
public:
	TRect(float aX=0, float aY=0, float aWidth=0, float aHeight=0) : iPosition(aX, aY), iSize(aWidth, aHeight) { }
	TRect(const TPoint& aPoint, const TSize& aSize) : iPosition(aPoint), iSize(aSize) { }
	TRect(const TPoint& aPoint) : TRect(aPoint, TSize(0, 0)) { }
	TRect(const TSize& aSize) : TRect(TPoint(0, 0), aSize) { }
	TRect(const SDL_Rect& aSdlRect) : TRect(TPoint(aSdlRect.x, aSdlRect.y), TSize(aSdlRect.w, aSdlRect.h)) { }
	TRect(const GPU_Rect& aGPURect) : TRect(TPoint(aGPURect.x, aGPURect.y), TSize(aGPURect.w, aGPURect.h)) { }
	
	operator SDL_Rect() const { SDL_Rect rect = {(int) iPosition.iX, (int) iPosition.iY, (int) iSize.iWidth, (int) iSize.iHeight}; return rect; }
	operator GPU_Rect() const { GPU_Rect rect = {iPosition.iX, iPosition.iY, iSize.iWidth, iSize.iHeight}; return rect; }

	float X() const 		{ return iPosition.iX; }
	float Y() const			{ return iPosition.iY; }
	float Width() const		{ return iSize.iWidth; }
	float Height() const	{ return iSize.iHeight; }
	
	const TPoint& Position() const 	{ return iPosition; }
	const TSize& Size() const 		{ return iSize; }

	void SetPosition(const TPoint& aPoint)	{ iPosition = aPoint; }
	void SetSize(const TSize& aSize) 		{ iSize = aSize; }
	void SetWidth(int aWidth)				{ iSize.iWidth = aWidth; }
	void SetHeight(int aHeight)				{ iSize.iHeight = aHeight; }
	
	TPoint GetTopRight()	const { return iPosition + TPoint(iSize.iWidth, 0); }
	TPoint GetTopLeft()		const { return iPosition; }
	TPoint GetBottomRight()	const { return iPosition + TPoint(iSize.iWidth, iSize.iHeight); }
	TPoint GetBottomLeft()	const { return iPosition + TPoint(0, iSize.iHeight); }
	TPoint GetCenter()		const { return iPosition + TPoint(iSize.iWidth/2, iSize.iHeight/2); }
	
	TRect& operator+=(const TPoint& aPoint) { iPosition += aPoint; return *this; }
	friend TRect operator+(TRect aRect, const TPoint& aPoint) { aRect.iPosition += aPoint; return aRect; }
	TRect& operator-=(const TPoint& aPoint) { iPosition -= aPoint; return *this; }
	friend TRect operator-(TRect aRect, const TPoint& aPoint) { aRect.iPosition -= aPoint; return aRect; }
	
	TRect& operator+=(const TSize& aSize) { iSize += aSize; return *this; }
	friend TRect operator+(TRect aRect, const TSize& aSize) { aRect.iSize += aSize; return aRect; }
	TRect& operator-=(const TSize& aSize) { iSize -= aSize; return *this; }
	friend TRect operator-(TRect aRect, const TSize& aSize) { aRect.iSize -= aSize; return aRect; }
	
	bool operator==(const TRect& aOther) const { return (iPosition == aOther.iPosition) && (iSize == aOther.iSize); }
	
	void Grow(const TSize& aSize) { iPosition.iX -= aSize.iWidth; iPosition.iY -= aSize.iHeight; iSize.iWidth += aSize.iWidth*2; iSize.iHeight += aSize.iHeight*2; }
	void Grow(float aDelta) { Grow(TSize(aDelta, aDelta)); }
	
	void ScaleKeepingAspectRatio(const TRect& aOuterRect);
	void Center(const TRect& aOuterRect) { AlignHCenter(aOuterRect); AlignVCenter(aOuterRect); }
	
	bool IsEmpty() const { return iSize.IsEmpty(); }
	bool HasPoint(const TPoint& aPoint) const { return (aPoint.iX > iPosition.iX) && (aPoint.iX < iPosition.iX + iSize.iWidth) && 
		(aPoint.iY > iPosition.iY) && (aPoint.iY < iPosition.iY + iSize.iHeight); }
		
	void AlignLeft(const TRect& aOuterRect) { iPosition.iX = aOuterRect.iPosition.iX; }
	void AlignRight(const TRect& aOuterRect) { float diff = aOuterRect.iSize.iWidth - iSize.iWidth ; iPosition.iX = aOuterRect.iPosition.iX + diff; }
	void AlignHCenter(const TRect& aOuterRect) { float diff = aOuterRect.iSize.iWidth - iSize.iWidth ; iPosition.iX = aOuterRect.iPosition.iX + diff/2; }
	
	void AlignTop(const TRect& aOuterRect) { iPosition.iY = aOuterRect.iPosition.iY; }
	void AlignBottom(const TRect& aOuterRect) { float diff = aOuterRect.iSize.iHeight - iSize.iHeight ; iPosition.iY = aOuterRect.iPosition.iY + diff; }
	void AlignVCenter(const TRect& aOuterRect) { float diff = aOuterRect.iSize.iHeight - iSize.iHeight ; iPosition.iY = aOuterRect.iPosition.iY + diff/2; }
	
	TRect& ApplyMargin(const TMargin& aMargin) { iPosition.iX += aMargin.iLeft; iPosition.iY += aMargin.iTop; 
		iSize.iWidth -= aMargin.iLeft + aMargin.iRight; iSize.iHeight -= aMargin.iTop + aMargin.iBottom; return *this;}
		
	void SetToMaxSquare() 
	{ 
		float delta = iSize.iWidth - iSize.iHeight; 
		if (delta>0) 
		{
			iSize.iWidth=iSize.iHeight; iPosition.iX += delta/2;
		}
		else 
		{
			iSize.iHeight=iSize.iWidth; iPosition.iY -= delta/2;
		}
	}

private:
	TPoint iPosition = {0, 0};
	TSize  iSize = {0, 0};
};


class TColor
{
public:
	static const int KUndefinedColor	= 0x80000000;
	static const int KRed				= 0xff0000;
	static const int KGreen				= 0x00ff00;
	static const int KBlue				= 0x0000ff;
	static const int KWhite				= 0xffffff;
	static const int KBlack				= 0x000000;
	
	TColor(int aColor = KBlack) : iColor(aColor) {}
	TColor(int aR, int aG, int aB) : iColor(MakeColor(aR, aG, aB)) { }
	
	operator SDL_Color() const {SDL_Color color = {(Uint8) ((iColor & 0xff0000) >> 16), (Uint8) ((iColor & 0xff00) >> 8), (Uint8) (iColor & 0xff), 0xff}; return color;}
	
	static int MakeColor(int aR, int aG, int aB) { return (aR << 16) | (aG << 8) | aB; }
	
	void Gradient(int aGradient);
	static void GradientComponent(int aGradient, Uint8& aComponent);
	
	Uint32 GetRGBAColor() { return (0xFF << 24)  | ((iColor & 0xff) << 16) |  (iColor & 0xff00) | ((iColor & 0xff0000) >> 16); }
	
	int GetA() const { return (iColor & 0xff000000) >> 24; }
	int GetR() const { return (iColor & 0xff0000) >> 16; }
	int GetG() const { return (iColor & 0xff00) >> 8; }
	int GetB() const { return (iColor & 0xff); }
	
	void OverrideDefaults(const TColor& aOther)
	{
		if (aOther.iColor != KUndefinedColor)
			iColor = aOther.iColor;
	}
	
	void SetDefault(const TColor& aOther)
	{
		if (iColor == KUndefinedColor)
			iColor = aOther.iColor;
	}
	
	int iColor = 0;
};


class TRoundedRect
{
public:
	TRoundedRect(int aRadiusX=-1, int aRadiusY=-1, TColor aFrameColor=TColor::KUndefinedColor, TColor aFillColor=TColor::KUndefinedColor, int aFrameWidth=-1) 
	: iRadiusX(aRadiusX), iRadiusY(aRadiusY), iFrameColor(aFrameColor), iFillColor(aFillColor), iFrameWidth(aFrameWidth) { }
	
	void OverrideDefaults(const TRoundedRect& aOther)
	{
		OverridePositiveValue(iRadiusX, aOther.iRadiusX);
		OverridePositiveValue(iRadiusY, aOther.iRadiusY);
		OverridePositiveValue(iFrameWidth, aOther.iFrameWidth);
		iFrameColor.OverrideDefaults(aOther.iFrameColor);
		iFillColor.OverrideDefaults(aOther.iFillColor);
	}
	
	int   	iRadiusX;
	int   	iRadiusY;
	TColor 	iFrameColor;
	TColor 	iFillColor;
	int		iFrameWidth;
};


class TFont
{
public:
	TFont(const std::string& aName = "", int aSize = -1, TColor aTextColor = TColor::KUndefinedColor) : iName(aName), iSize(aSize), iTextColor(aTextColor) {}
	void OverrideDefaults(const TFont& aOther)
	{
		OverrideNonEmptyString(iName, aOther.iName);
		OverridePositiveValue(iSize, aOther.iSize);
		OverridePositiveValue(iOutline, aOther.iOutline);
		iTextColor.OverrideDefaults(aOther.iTextColor);
		iBold = aOther.iBold;
		iItalic = aOther.iItalic;
		iUnderline = aOther.iUnderline;
		iStrikeThrough = aOther.iStrikeThrough;
	}

	void SetDefaults(const TFont& aOther)
	{
		SetDefaultString(iName, aOther.iName);
		SetDefaultPositiveValue(iSize, aOther.iSize);
		SetDefaultPositiveValue(iOutline, aOther.iOutline);
		iTextColor.SetDefault(aOther.iTextColor);
		iBold = aOther.iBold;
		iItalic = aOther.iItalic;
		iUnderline = aOther.iUnderline;
		iStrikeThrough = aOther.iStrikeThrough;
	}
	
	std::string	iName = "";
	int			iSize = -1;
	int 		iOutline = 0;
	TColor		iTextColor = TColor::KUndefinedColor;
	bool 		iBold = false;
	bool 		iItalic = false;
	bool 		iUnderline = false;
	bool 		iStrikeThrough = false;
};


/*********************************************
	* 
	* TPerformance
	*
	*********************************************/

class TPerformance
{
public:
	TPerformance(const std::string& aText) : iText(aText)
	{
		if (sLevel == 0)
		{
			printf("\n");
		}
		
		printf("%s: start\n", iText.c_str());
		iLevel = sLevel++;
	}
	
	~TPerformance()
	{
		sLevel--;
		printf("%s: done, total=%d\n", iText.c_str(), SDL_GetTicks() - iStart);
	}
	
	void TimeStamp(const std::string& aText)
	{
		Uint32 now = SDL_GetTicks();
		
		if (iLast == 0)
			printf("%s, total=%d\n", aText.c_str(), now - iStart);
		else
			printf("%s, delta=%d, accumulated=%d\n", aText.c_str(), now - iLast, now - iStart);
		
		iLast = now;
	}
	
private:
	int iLevel;
	static int sLevel;
	Uint32 iLast = 0;
	Uint32 iStart = SDL_GetTicks();
	std::string iText;
};
	
} // namespace sdl

#endif // __SDL_BASE_H