/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlcontainer.h"
#include "sdllayout.h"

using namespace sdl;


Container::~Container()
{
	RemoveAllWidgets();
}

void Container::RemoveWidget(Widget& aWidget)
{
	iWidgets.erase(std::remove(iWidgets.begin(), iWidgets.end(), &aWidget), iWidgets.end()); 
	aWidget.SetParent(nullptr); 
	if (aWidget.GetParentOwnsMe())
		delete &aWidget;
}

void Container::RemoveAllWidgets()
{
	for (auto &widget : iWidgets)
	{
		widget->SetParent(nullptr); 
		if (widget->GetParentOwnsMe())
			delete widget;
	}
	iWidgets.clear();
}

void Container::LayoutChidlren()
{
	for (auto &widget : iWidgets)
		widget->Layout();
}

void Container::LayoutSingleChild(Widget& aWidget)
{
	TRect rect = TRect(GetClientAreaSize());
	aWidget.SetRect(rect);
	aWidget.Layout();
}

void Container::Layout()
{
	if (iWidgets.size() == 1)
		LayoutSingleChild(*iWidgets[0]);
	else
		LayoutChidlren();
}

void Container::Render(IRenderer& aRenderer)
{
	if (!IsEnabled())
		return;
	
	Widget::Render(aRenderer);
	
	for (auto &widget : iWidgets)
		widget->Render(aRenderer);
}

void Container::ResetSDLState() 
{ 
	Widget::ResetSDLState();

	for (auto &widget : iWidgets)
		widget->ResetSDLState();
}

void Container::SizeChanged(const TSize& aOldSize)
{
	for (auto &widget : iWidgets)
		widget->SizeChanged(aOldSize);

	Widget::SizeChanged(aOldSize);
}

void Container::PositionChanged(const TPoint& aOldPosition)
{
	Widget::PositionChanged(aOldPosition);
}


void Container::ResizeToClientArea(const TSize& aClientAreaSize)
{
	TSize containerSize = GetSize();
	TSize clientRectSize = GetClientAreaSize();
	
	SetSize(aClientAreaSize + containerSize - clientRectSize);
}

Widget* Container::FindFocusNext()
{
	Widget* widget = nullptr;

	// reset if out of range (important when wrapping around)
	if (!(iFocussedWidgetIndex>=0 && iFocussedWidgetIndex<(int)iWidgets.size()))
	{
		HightlightChanged(THighlightChange::ETop);
		iFocussedWidgetIndex = 0;
	}
	
	int oldIndex = iFocussedWidgetIndex;

	bool hasWrappedAround = false;
	
	while ((iFocussedWidgetIndex>=0 && iFocussedWidgetIndex<(int)iWidgets.size()))
	{
		// 1. check if can focus widget - if so done
		widget = iWidgets[iFocussedWidgetIndex];
		if (widget->HasInput() && !widget->HasFocus())
			break;
		
		// 2. if not, check if widget can move focus (if it's a container). If so, done
		widget = widget->FindFocusNext();
		if (widget)
			break;

		// 3. Update for next round
		iFocussedWidgetIndex++;
		
		// 4. If at end of the list, we'll wrap around if we are the top level widget
		// otherwise we're done
		if (iFocussedWidgetIndex == (int) iWidgets.size())
		{
			if (GetParent() == nullptr)
			{
				iFocussedWidgetIndex = 0;
			}
			else
			{
				break;
			}
		}
		
		// 5. If we've wrapped around and back at where we started, we're done.
		//    Only exception is when we are the top-level widget and we have only 1 child with focussable items. 
		//    When we ask that child to switch focus it won't wrap around when the focus was at the beginning or end of the 
		//    list (which is correct behavior). In order for that child to do wrap around we need to ask it again. And hence
		//    we need to make sure top-level widgets wrap around at least once.
		if ((oldIndex == iFocussedWidgetIndex) && (GetParent() == nullptr))
		{
			if (hasWrappedAround)
			{
				widget = nullptr;
				break;
			}
			hasWrappedAround = true;
		}
	}
	
	if (widget)
	{
		HightlightChanged(THighlightChange::EDown);
	}
	
	return widget;
}

// same structure (and therefore same comments apply) as in FindFocusNext
Widget* Container::FindFocusPrevious()
{
	Widget* widget = nullptr;
	
	if (!(iFocussedWidgetIndex>=0 && iFocussedWidgetIndex<(int)iWidgets.size()))
	{
		HightlightChanged(THighlightChange::EBottom);
		iFocussedWidgetIndex = iWidgets.size()-1;
	}

	int oldIndex = iFocussedWidgetIndex;
	
	bool hasWrappedAround = false;
	
	while ((iFocussedWidgetIndex>=0 && iFocussedWidgetIndex<(int)iWidgets.size()))
	{
		widget = iWidgets[iFocussedWidgetIndex];
		
		if (widget->HasInput() && !widget->HasFocus())
			break;
		
		widget = widget->FindFocusPrevious();
		if (widget)
			break;
		
		iFocussedWidgetIndex--;
		if (iFocussedWidgetIndex == -1)
		{
			if (GetParent() == nullptr)
			{
				iFocussedWidgetIndex = (int) iWidgets.size()-1;
			}
			else
			{
				break;
			}
		}
		
		if ((oldIndex == iFocussedWidgetIndex) && (GetParent() == nullptr))
		{
			if (hasWrappedAround)
			{
				widget = nullptr;
				break;
			}
			hasWrappedAround = true;
		}
	}
	
	if (widget)
	{
		HightlightChanged(THighlightChange::EUp);
	}
	
	return widget;
}

bool Container::SwitchFocusNext()
{
	return HasInput() ? SwitchFocus(FindFocusNext()) : false;
}

bool Container::SwitchFocusPrevious()
{
	return HasInput() ? SwitchFocus(FindFocusPrevious()) : false;
}


bool Container::SwitchFocus(Widget* aWidget)
{
	if (aWidget)
	{
		if (iCurrentFocussedWidget)
			iCurrentFocussedWidget->SetFocus(false);
		iCurrentFocussedWidget = aWidget;
		iCurrentFocussedWidget->SetFocus(true);
	}
	return aWidget != nullptr;
}


bool Container::DelegateCommandToChildren(TCommand aCommand)
{
	bool consumed = false;
	for (auto &widget: iWidgets)
	{
		consumed = widget->HandleCommand(aCommand);
		if (consumed)
			break;
	}
	return consumed;
}


bool Container::HandleCommand(TCommand aCommand)
{
	bool consumed = DelegateCommandToFocussedWidget(aCommand);
	if (!consumed)
		consumed = HandleCommandMyself(aCommand);
	if (!consumed)
		consumed = DelegateCommandToChildren(aCommand);
	
	if (consumed)
		System::RequestReRender();
	
	return consumed;
}

bool Container::DelegateCommandToFocussedWidget(TCommand aCommand)
{
	return iCurrentFocussedWidget ? iCurrentFocussedWidget->HandleCommand(aCommand) : false;
}


bool Container::HandleCommandMyself(TCommand aCommand)
{
	bool consumed = false;
	
	switch (aCommand)
	{
		case TCommand::EDown:
		case TCommand::EFocusNext:
			consumed = SwitchFocusNext();
			break;
			
		case TCommand::EUp:
		case TCommand::EFocusPrevious:
			consumed = SwitchFocusPrevious();
			break;
			
		default:
			break;
	}
	return consumed;
}
