/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlimagewidget.h"

using namespace sdl;

template<class T>
class GlobalImageCache
{
public:
	GlobalImageCache(int aMaxSize = 16) : iMaxSize(aMaxSize) {}
	~GlobalImageCache() { Flush(); }
	std::shared_ptr<T> Add(const std::string& aFileName, T* aImage, const TSize& aSize = TSize(0, 0));
	std::shared_ptr<T> Get(const std::string& aFileName, const TSize& aSize = TSize(0, 0));
	void Flush();
private:
	void Evict();
	void Stats();
	std::string GetCacheEntryName(const std::string& aFileName, const TSize& aSize);
	std::map<std::string, std::shared_ptr<T>> iCache;
	std::map<std::string, int> iObjectAge;
	int iMaxSize;
	int iCount = 0;
	
	int iLookups=0;
	int iHits=0;
};

GlobalImageCache<ITexture> sTextureCache;


template<class T>
void GlobalImageCache<T>::Flush()
{
	iCache.clear();
	iObjectAge.clear();
}

template<class T>
void GlobalImageCache<T>::Evict()
{
	int minAge = iCount;
	std::string entryName;
	for (auto &entry : iObjectAge)
	{
		if (entry.second < minAge)
		{
			minAge = entry.second;
			entryName = entry.first;
		}
	}
	if (entryName.size() > 0)
	{
//		printf("GlobalImageCache::Evict %s\n", entryName.c_str());
		iObjectAge.erase(entryName);
		iCache.erase(entryName);
	}
}

template<class T> void DestroyImage(T* aImage);

template<>
void DestroyImage(ITexture* aTexture)
{
//	printf("DestroyImage 0x%p\n", aTexture);
	delete(aTexture);
}

template<class T>
std::shared_ptr<T> GlobalImageCache<T>::Add(const std::string& aFileName, T* aImage, const TSize& aSize)
{
	if (aImage == nullptr)
		return std::shared_ptr<T>(nullptr);
	
	std::string entryName = GetCacheEntryName(aFileName, aSize);
	
//	printf("GlobalImageCache::Add %s --> %p\n", entryName.c_str(), &aTexture);
	
	iCache[entryName] = std::shared_ptr<T>(aImage, DestroyImage<T>);
	iObjectAge[entryName] = ++iCount;
	if ((int) iCache.size() > iMaxSize)
		Evict();
	
	return iCache[entryName];
}

template<class T>
std::shared_ptr<T> GlobalImageCache<T>::Get(const std::string& aFileName, const TSize& aSize)
{
	iLookups++;
	std::string entryName = GetCacheEntryName(aFileName, aSize);
	if (iCache.find(entryName) != iCache.end())
	{
		iHits++;
		Stats();
		iObjectAge[entryName] = ++iCount; // make recent item
		return iCache.at(entryName);
	}

	Stats();
	
	return std::shared_ptr<T>(nullptr);
}

template<class T>
void GlobalImageCache<T>::Stats()
{
//	printf("CACHE: count=%d, lookup=%d, hits=%d, ratio=%d%%\n", (int) iCache.size(), iLookups, iHits, (iHits * 100)/iLookups);
}


template<class T>
std::string GlobalImageCache<T>::GetCacheEntryName(const std::string& aFileName, const TSize& aSize)
{
	if ((aSize.iWidth > 0) && (aSize.iHeight > 0))
		return aFileName + "@" + std::to_string(aSize.iWidth) + "x" + std::to_string(aSize.iHeight);
	return aFileName;
}


void ImageWidget::SetDrawMode(const std::string& aDrawMode)
{
	if (aDrawMode.compare("normal") == 0)
		SetDrawMode(TDrawMode::ENormal);
	else if(aDrawMode.compare("scaled") == 0)
		SetDrawMode(TDrawMode::EScaled);
}

void ImageWidget::SetDrawMode(TDrawMode aDrawMode)
{
	iDrawMode = aDrawMode; 
	if (iDrawMode == TDrawMode::EScaled)
	{
		SetAlignment(Widget::TVerticalAlignment::ECenter);
		SetAlignment(Widget::THorizontalAlignment::ECenter);
	}
	else if (iDrawMode == TDrawMode::ENormal)
	{
		SetAlignment(Widget::TVerticalAlignment::ETop);
		SetAlignment(Widget::THorizontalAlignment::ELeft);
	}
	
	CacheClear(); 
}


void ImageWidget::OverrideDefaults(const Properties& aOther)
{
	iImageProperties.iRect = aOther.iRect;
	OverrideNonEmptyString(iImageProperties.iDrawMode, aOther.iDrawMode);
	ApplyProperties();
}



void ImageWidget::ResetSDLState()
{
	Widget::ResetSDLState(); 
	
	sTextureCache.Flush();
}

std::shared_ptr<ITexture> ImageWidget::GetSharedImage(IRenderer& aRenderer)
{
	std::shared_ptr<ITexture> sharedImage = sTextureCache.Get(iFileName, TSize(0, 0));
	if (sharedImage.get())
		return sharedImage;
	return sTextureCache.Add(iFileName, RenderUtils::CreateTextureFromFile(aRenderer, iFileName.c_str()));
}



void ImageWidget::Render(IRenderer& aRenderer)
{
	if (IsEnabled() && iFileName.size() > 3)
	{
		switch (iDrawMode)
		{
			case TDrawMode::EScaled:
			{
				std::shared_ptr<ITexture> sharedTexture(GetSharedImage(aRenderer));
				if (sharedTexture.get())
				{
					TRect boundingRect = GetScreenRect();
					boundingRect.ApplyMargin(iMargin);
					
					RenderUtils::RenderScaledTextureMaintainAspectRatio(aRenderer, *sharedTexture.get(), boundingRect);
				}
				break;
			}
			
			case TDrawMode::ENormal:
			default:
			{
				std::shared_ptr<ITexture> sharedTexture = GetSharedImage(aRenderer);
				if (sharedTexture.get())
				{
					TRect boundingRect = GetScreenRect();
					boundingRect.ApplyMargin(iMargin);
					aRenderer.RenderTexture(*sharedTexture.get(), boundingRect.Position());
				}
				break;
			}
		}
	}
	Widget::Render(aRenderer);
}
