/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdljoy.h"

using namespace sdl;

namespace sdl
{
	Uint32 JoyCallback(Uint32 aInterval, void* aParam)
	{
		JoystickHatEventHandler* self = (JoystickHatEventHandler*) aParam;
		self->RepeatEvents();
		
		return aInterval;
	}
}

Joystick::Joystick()
{
	EventManager::SetJoystickHatEventHandler(&iJoystickHatEventHandler);
	EventManager::SetJoystickButtonEventHandler(&iJoystickButtonEventHandler);
}

void Joystick::Close()
{
	for (auto &sdlJoysyick : iSdlJoysticks)
		SDL_JoystickClose(sdlJoysyick);
	iSdlJoysticks.clear();
}

void Joystick::Open()
{
	int numJoysticks = SDL_NumJoysticks();
	
	printf("SDL: number of joysticks=%d\n", numJoysticks);
	
	for (int i=0 ; i<numJoysticks ; i++)
	{
		SDL_Joystick* sdlJoystick = SDL_JoystickOpen(i);
		if (sdlJoystick)
		{
			iSdlJoysticks.push_back(sdlJoystick);
			printf("SDL Joystick %d\n", i);
			printf("  Name: %s\n", SDL_JoystickNameForIndex(i));
			printf("  Number of Axes: %d\n", SDL_JoystickNumAxes(sdlJoystick));
			printf("  Number of Buttons: %d\n", SDL_JoystickNumButtons(sdlJoystick));
			printf("  Number of Balls: %d\n", SDL_JoystickNumBalls(sdlJoystick));
			printf("  Number of Hats: %d\n", SDL_JoystickNumHats(sdlJoystick));
		}
		else
		{
			printf("SDL_JoystickOpen(%d) failed: %s\n", i, SDL_GetError());
		}
	}
}


static SDL_TimerID		sTimerId = 0;


JoystickHatEventHandler::~JoystickHatEventHandler()
{
	StopGeneratingEvents();
}


void JoystickHatEventHandler::RepeatEvents()
{
	if (IsHatDown(iEventHatCode))
		GenerateCommand(TCommand::EDown);
	
	if (IsHatUp(iEventHatCode))
		GenerateCommand(TCommand::EUp);
	
	if (iScroller.ShouldFastScrollingBeStarted())
	{
		StopGeneratingEvents();
		StartGeneratingEvents(true);
	}
}

void JoystickHatEventHandler::GenerateCommand(TCommand aCommand)
{
	EventManager::InjectCustomCommand(aCommand);
}


void JoystickHatEventHandler::StartGeneratingEvents(bool aFastScrolling)
{
	if (sTimerId == 0)
	{
		if (!aFastScrolling)
			iScroller.Start();
		sTimerId = SDL_AddTimer(aFastScrolling ? SDLJoystickProperties.Navigation.Speed.iFastScrollSpeedInMs : SDLJoystickProperties.Navigation.Speed.iScrollSpeedInMs, JoyCallback, this);
	}
}

void JoystickHatEventHandler::StopGeneratingEvents()
{
	if (sTimerId != 0)
	{
		SDL_RemoveTimer(sTimerId);
		sTimerId = 0;
	}
}


void JoystickHatEventHandler::ProcessJoyHatEvent(const SDL_JoyHatEvent& aEvent)
{
	int oldHatDown = iEventHatCode;
	iEventHatCode = aEvent.value;
	if (aEvent.which == SDLJoystickProperties.Navigation.iJoystick)
	{
		// generate up/down/left/right command (if not done previously)
		if (!IsHatDown(oldHatDown) && IsHatDown(iEventHatCode))
			GenerateCommand(TCommand::EDown);
		
		if (!IsHatUp(oldHatDown) && IsHatUp(iEventHatCode))
			GenerateCommand(TCommand::EUp);
		
		if (!IsHatLeft(oldHatDown) && IsHatLeft(iEventHatCode))
			GenerateCommand(TCommand::ELeft);
		
		if (!IsHatRight(oldHatDown) && IsHatRight(iEventHatCode))
			GenerateCommand(TCommand::ERight);
		
		// repeat up & down events
		if (IsHatUp(iEventHatCode) || IsHatDown(iEventHatCode))
			StartGeneratingEvents();
		else
			StopGeneratingEvents();
	}
}


void JoystickButtonEventHandler::ProcessJoyButtonEvent(const SDL_JoyButtonEvent& aEvent)
{
	if (aEvent.state == SDL_PRESSED)
	{
		for (auto &it : SDLJoystickProperties.iButtons)
		{
			if ((it.iJoystickId == aEvent.which) && (it.iJoystickButtonId == aEvent.button))
			{
				EventManager::InjectCustomCommand((TCommand) it.iCommand);
				break;
			}
		}
	}
}

