/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlmenu.h"

using namespace sdl;

bool myCompareFunction(const std::unique_ptr<MenuItem>& a, const std::unique_ptr<MenuItem>& b) { return a.get()->GetName() < b.get()->GetName(); }

void Menu::Sort()
{
	std::sort(iMenuItems.begin(), iMenuItems.end(), myCompareFunction);
}

const Menu& MenuItem::GetParentMenu() const
{
	static Menu sMenu(""); 
	return iParentMenu ? *iParentMenu : sMenu;
}


void Menu::AddMenuItem(MenuItem* aMenuItem, TAddPolicy aAddPolicy, TNotifiyPolicy aNotifiyPolicy)
{
	aMenuItem->SetParent(this);
	if (aAddPolicy == TAddPolicy::EAddToFront)
		iMenuItems.insert(iMenuItems.begin(), std::unique_ptr<MenuItem>(aMenuItem));
	else
		iMenuItems.push_back(std::unique_ptr<MenuItem>(aMenuItem));
	
	if (aNotifiyPolicy == TNotifiyPolicy::ENotifyClients)
		MenuListChanged();
}

void Menu::AddMenu(Menu* aMenu, TAddPolicy aAddPolicy)
{
	AddMenuItem(aMenu, aAddPolicy);
	MenuListChanged();
}

const MenuItem& Menu::GetMenuItem(int aPosition) const
{
	static MenuItem KEmptyMenuItem("", "");
	if ((aPosition > (int) iMenuItems.size() - 1) || (aPosition < 0))
		return KEmptyMenuItem;
	return *(iMenuItems.at(aPosition));
}

MenuItem& Menu::GetMenuItem(int aPosition)
{
	static MenuItem KEmptyMenuItem("", "");
	if ((aPosition > (int) iMenuItems.size() - 1) || (aPosition < 0))
	{
		return KEmptyMenuItem;
	}
	return *(iMenuItems.at(aPosition));
}


void Menu::SetHighlightedItem(int aPosition)
{
	if (aPosition>=0 && aPosition<(int)iMenuItems.size())
	{
		iPosition = aPosition;
		MenuItemHighlighted();
	}
}

bool Menu::ScrollUp()
{
	if (iPosition > 0)
	{
		iAtTop = iAtBottom = false;
		SetHighlightedItem(iPosition-1);
	} 
	else
	{
		iAtTop = true;
	}
	
	return iAtTop;
}

bool Menu::ScrollDown()
{
	int bottom = MenuSize() - 1;
	if (iPosition < bottom) 
	{
		iAtTop = iAtBottom = false;
		SetHighlightedItem(iPosition+1);
	}
	else
	{
		iAtBottom = true;
	}

	return iAtBottom;
}

bool Menu::ScrollPageUp()
{
	if (iPosition > 0)
	{
		iAtTop = iAtBottom = false;

		int newPosition = FindPrevMenuItemByStartChar(GetStartCharMenuItem(iPosition));
		if (newPosition == -1)
			newPosition = 0;
		SetHighlightedItem(newPosition);
	}
	else
	{
		iAtTop = true;
	}
	
	return iAtTop;
}

bool Menu::ScrollPageDown()
{
	if (iPosition < (int)iMenuItems.size() - 1)
	{
		iAtTop = iAtBottom = false;

		int newPosition = FindNextMenuItemByStartChar(GetStartCharMenuItem(iPosition));
		if (newPosition == -1)
			newPosition = iMenuItems.size() - 1;
		SetHighlightedItem(newPosition);
	}
	else
	{
		iAtBottom = true;
	}

	return iAtBottom;
}


void Menu::KeepInRange()
{
	if (iPosition <= 0)
		iPosition = 0;
	else if (iPosition >= MenuSize())
		iPosition = MenuSize() - 1;
}


void Menu::Activate()
{
	iIsActive = true;
	ResetTopAndBottom();
	MenuItemHighlighted();
}


bool Menu::HasMenuItem(const std::string& aCommand)
{
	for (auto &menuItem : iMenuItems)
	{
		if (menuItem.get()->GetCommand() == aCommand)
			return true;
	}
	return false;
}


void Menu::DeleteMenuItem(const std::string& aCommand, TNotifiyPolicy aNotifiyPolicy)
{
	bool changed = false;
	for (auto it = iMenuItems.begin() ; it != iMenuItems.end() ; ++it)
	{
		if ((*it)->GetCommand() == aCommand)
		{
			iMenuItems.erase(it);
			KeepInRange();
			changed = true;
			break;
		}
	}
	
	if (changed && (aNotifiyPolicy == TNotifiyPolicy::ENotifyClients))
		MenuListChanged();
}

void Menu::DeleteAllMenuItems()
{
	iMenuItems.clear();
	iPosition = 0;
	MenuListChanged();
}

void Menu::PruneMenuItems(int aMax, TNotifiyPolicy aNotifiyPolicy)
{
	int diff = iMenuItems.size() - aMax;
	if (diff > 0)
	{
		iMenuItems.erase(iMenuItems.end() - diff, iMenuItems.end());
		iPosition = 0;
		if (aNotifiyPolicy == TNotifiyPolicy::ENotifyClients)
			MenuListChanged();
	}
}

void Menu::AddMenuItems(std::vector<std::string>& aList)
{
	for (auto &item : aList)
		AddMenuItem(new MenuItem(item));
	MenuListChanged();
}

void Menu::AddMenuItemListener(IMenuItemListener& aMenuItemListener, const std::string& aMenuType)
{
	if ((aMenuType.size()==0) || (GetType().compare(aMenuType) == 0))
	{
		iMenuItemListeners.insert(&aMenuItemListener);

		const MenuItem& currentMenuItem = GetCurrentMenuItem();
		if (!currentMenuItem.GetCommand().empty() && iIsActive)
			aMenuItemListener.MenuItemHighlighted(currentMenuItem);
	}
	
	for (auto &menuItem : iMenuItems)
	{
		Menu* menu = menuItem.get()->GetSubMenu();
		if (menu)
			menu->AddMenuItemListener(aMenuItemListener, aMenuType);
	}
}


void Menu::MenuItemHighlighted()
{
	if (iIsActive)
	{
		const MenuItem& currentMenuItem = GetCurrentMenuItem();
		for (auto &menuItemListener : iMenuItemListeners)
			menuItemListener->MenuItemHighlighted(currentMenuItem);
	}
}

void Menu::MenuItemCommand(const TCommand& aCommand)
{
	// take copy of menu item (rather than a reference) - one of the command may be to delete the item
	// if we don't take a copy we end up in a stale reference
	const MenuItem currentMenuItem = GetCurrentMenuItem();
	for (auto &menuItemListener : iMenuItemListeners)
		menuItemListener->MenuItemCommand(currentMenuItem, aCommand);
}

int Menu::FindMenuItemPosition(const std::string& aMenuItem)
{
	int i = 0;
	for (auto &menuItem : iMenuItems)
	{
		if (menuItem.get()->GetName().compare(aMenuItem) == 0)
			return i;
		i++;
	}
	return -1;
}


void Menu::HighlightItem(const std::string& aMenuItem)
{
	SetHighlightedItem(FindMenuItemPosition(aMenuItem));
}


Menu* Menu::FindMenu(const std::string& aName)
{
	if (GetName().compare(aName) == 0)
		return this;
	
	for (auto &menuItem : iMenuItems)
	{
		Menu* menu = menuItem.get()->GetSubMenu();
		if (menu)
		{
			menu = menu->FindMenu(aName);
			if (menu)
				return menu;
		}
	}
	
	return nullptr;
}

void Menu::MenuListChanged()
{
	if (IsActive())
	{
		for (auto &menuItemListener : iMenuItemListeners)
			menuItemListener->MenuListChanged(*this);
	}
}

unsigned char Menu::GetStartCharMenuItem(int aPosition)
{
	unsigned char startChar = 0;
	if (iMenuItems.size() > 0)
	{
		startChar = iMenuItems.at(aPosition)->GetName().at(0);
		if (startChar>='0' && startChar<='9')
			startChar = '0';
		
	}
	return startChar;
}


int Menu::FindNextMenuItemByStartChar(unsigned char aChar)
{
	for (int i=iPosition+1 ; i<(int)iMenuItems.size() ; i++)
	{
		char startChar = GetStartCharMenuItem(i);
		if (startChar > aChar)
			return i;
	}
	return -1;
}

int Menu::FindPrevMenuItemByStartChar(unsigned char aChar)
{
	for (int i=iPosition-1 ; i>=0 ; i--)
	{
		char startChar = GetStartCharMenuItem(i);
		if (startChar < aChar)
			return i;
	}
	return -1;
}
