/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlmenu2dmodel.h"

using namespace sdl;


/********************************
 * 
 * Menu2DModel
 * 
 ********************************/

static Menu	KEmptyMenu("");

Menu* Menu2DModel::ConstructMenu(const std::string& aName, const std::string& aType, const TStringPairList& aStringPairList)
{
	Menu* menu = new Menu(aName.c_str(), aType.c_str());
	for (auto &stringPair : aStringPairList) 
		menu->AddMenuItem(new MenuItem(stringPair.second.c_str(), stringPair.first.c_str()));
	
	return menu;
}


Menu& Menu2DModel::CurrentMenu()
{
	if (iSubMenu)
		return *iSubMenu;

	return KEmptyMenu;
}
	
bool Menu2DModel::DoSwitchMenu(int aDelta)
{
	bool switched = false;
	int oldMenu = iCurrentMenu;
	if (iMenus.size() > 1)
	{
		while (true)
		{
			iCurrentMenu += aDelta;
			
			if (iCurrentMenu >= (int) iMenus.size())
				iCurrentMenu = 0;
			if (iCurrentMenu < 0)
				iCurrentMenu = iMenus.size()-1;
			
			if (iCurrentMenu == oldMenu)
				break;
			
			if (iMenus[iCurrentMenu].get()->IsEnabled())
			{
				ActivateMenu(iMenus[iCurrentMenu].get());
				switched = true;
				
				break;
			}
		}

	}

	return switched;
}

void Menu2DModel::AddMenu(Menu* aMenu)
{
	if(aMenu)
	{
		iMenus.push_back(std::unique_ptr<Menu>(aMenu));
		if (iMenus.size() == 1)
		{
			iCurrentMenu = 0;
			ActivateMenu(aMenu);
		}
		
		for (auto &menuItemListener : iMenuItemListeners[aMenu->GetType()])
			aMenu->AddMenuItemListener(*menuItemListener, aMenu->GetType());
		for (auto &menuItemListener : iMenuItemListeners[""])
			aMenu->AddMenuItemListener(*menuItemListener, aMenu->GetType());
	}
}

Menu* Menu2DModel::ActivateMenu(const std::string& aMenuName)
{
	int menuItemIndex;
	Menu* menu = FindMenu(aMenuName, menuItemIndex);
	if (menu)
	{
		iCurrentMenu = menuItemIndex;
		ActivateMenu(menu);
	}
	return menu;
}

bool Menu2DModel::SelectItem()
{
	bool consumed = false;

	MenuItem& currenMenutItem = CurrentMenu().GetCurrentMenuItem();
	Menu* subMenu = currenMenutItem.GetSubMenu();
	if (subMenu)
	{
		ActivateMenu(subMenu);
		consumed = true;
	}

	return consumed;
}

bool Menu2DModel::GoBack()
{
	bool consumed = false;
	Menu* subMenu = CurrentMenu().GetParentMenu();
	if (subMenu)
	{
		ActivateMenu(subMenu);
		consumed = true;
	}
	return consumed;
}

void Menu2DModel::ActivateMenu(Menu* aMenu)
{
	if (iSubMenu)
		iSubMenu->DeActivate();
	
	iSubMenu = aMenu;
	MenuChanged();
	iSubMenu->Activate();
}

void Menu2DModel::DeleteMenusByType(const std::string& aType)
{
	std::string oldMenuName = CurrentMenu().GetName();

	// remove & delete the menus
	iMenus.erase(std::remove_if(iMenus.begin(), iMenus.end(), [=](std::unique_ptr<Menu>& aMenu) { return aMenu.get()->GetType() == aType; } ), iMenus.end());

	// find index current menu (if still in the menu list)
	int index = 0;
	for (auto &menu : iMenus)
	{
		if (menu->GetName() == oldMenuName)
			break;
		index++;
	}
	// not found
	if (index == (int) iMenus.size())
		index = -1;

	// activate new menu if needed
	iCurrentMenu = index;
	if (iCurrentMenu != -1)
		ActivateMenu(iMenus[iCurrentMenu].get());
}

void Menu2DModel::EnableMenu(const std::string& aName, bool aEnable)
{
	auto menuIt = std::find_if(iMenus.begin(), iMenus.end(), [=](std::unique_ptr<Menu>& aMenu) { return (aMenu.get()->GetName()) == aName; } );
	if (menuIt != iMenus.end())
		(*menuIt).get()->SetEnabled(aEnable);
}

void Menu2DModel::GetMenuList(const Menu& aMenu, TStringPairList& aStringPairList)
{
	for (auto &menuItem : aMenu.GetMenuItems())
		aStringPairList.push_back(std::make_pair(menuItem.get()->GetCommand(), menuItem.get()->GetName()));
}

void Menu2DModel::AddMenuItemListener(IMenuItemListener& aMenuItemListener, const std::string& aMenuType)
{
	iMenuItemListeners[aMenuType].insert(&aMenuItemListener);
	for(auto &menu :iMenus)
		menu.get()->AddMenuItemListener(aMenuItemListener, aMenuType);
}

void Menu2DModel::AddMenuChangedListener(IMenuChangedListener& aMenuChangedListener)
{
	iMenuChangedListeners.push_back(&aMenuChangedListener);
	aMenuChangedListener.MenuChanged(CurrentMenu());
}

void Menu2DModel::MenuChanged()
{
	for (auto &menuChangedListener : iMenuChangedListeners)
		menuChangedListener->MenuChanged(CurrentMenu());
}

Menu* Menu2DModel::FindMenu(const std::string& aMenuName, int& aMenuIndex)
{
	int nrMenus = iMenus.size();
	for (aMenuIndex=0 ; aMenuIndex<nrMenus ; aMenuIndex++)
	{
		Menu* menu = iMenus[aMenuIndex].get()->FindMenu(aMenuName);
		if (menu)
			return menu;
	}
	return nullptr;
}


void Menu2DModel::HighlightMenuItem(const std::string& aMenuName, const std::string& aMenuItem)
{
	Menu* menu= ActivateMenu(aMenuName);
	if (menu)
		menu->HighlightItem(aMenuItem);
}

void Menu2DModel::GetHighlightMenuItem(std::string& aMenuName, std::string& aMenuItem)
{
	aMenuName = CurrentMenu().GetName();
	aMenuItem = CurrentMenu().GetCurrentMenuItem().GetName();
}
