/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlmenu2dwidget.h"
#include "sdluserfeedbackmanager.h"
#include "sdlthemeproperties.h"

using namespace sdl;

void MenuItemContainer::Populate(const Menu& aMenu)
{
	int halfItems = iNoOfMenuItems/2;
	int startMenuIndex = aMenu.GetCurrentMenuPosition() - halfItems; 
	for (int i=0 ; i<iNoOfMenuItems ; i++ )
		SetMenuItem(i, aMenu.GetMenuItem(startMenuIndex + i));
}

MenuItemWidget::MenuItemWidget() 
{
	iImageWidget.SetDrawMode(ImageWidget::TDrawMode::EScaled);
	iImageWidget.SetImageFile("menu.png");
	iImageWidget.SetEnabled(false);
	AddWidget(iLabelWidget);
	AddWidget(iImageWidget);
}

void MenuItemWidget::SetMenuText(const std::string& aMenuText, bool aIsSubMenu) 
{
	iLabelWidget.SetText(aMenuText);
	iImageWidget.SetEnabled(aIsSubMenu);
	SetImagePosition();
}

void MenuItemWidget::SetImagePosition()
{
//	TSize size = iLabelWidget.GetTextSize();
	int deltaY = (iLabelWidget.GetHeight() - iImageWidget.GetHeight()) / 2;
//	iImageWidget.SetPosition(TPoint(size.iWidth, 0) + TPoint(10, deltaY));

//	TPoint p = iLabelWidget.GetRenderedRect().GetTopRight();
//	printf("SETSIZE: labelTextSize=(%d, %d), RenderTextSize=(%d, %d)\n", size.iWidth, size.iHeight, p.iX, 0);
	iImageWidget.SetPosition(TPoint(iLabelWidget.GetRenderedRect().GetTopRight().iX, 0) + TPoint(10, deltaY));
}

void MenuItemWidget::SizeChanged(const TSize& aOldSize)
{
	iLabelWidget.SetSize(GetSize());
}

/*********************************************
 * 
 * TextListMenu
 * 
 *********************************************/


void TextListMenu::SetTextStyle(const std::string& aTextStyle)
{
	iHorizontalAlignement = Widget::THorizontalAlignment::ELeft;
	
	if (aTextStyle.compare("left") == 0)
		iTextStyle = TTextStyle::ELeftAligned;
	else if (aTextStyle.compare("centered") == 0)
	{
		iTextStyle = TTextStyle::ECentered;
		iHorizontalAlignement = Widget::THorizontalAlignment::ECenter;
	}
	else if (aTextStyle.compare("round") == 0)
		iTextStyle = TTextStyle::ERound;
}

void TextListMenu::SetProperties()
{
	TRect rect = iTextListMenuProperties.iArea.IsEmpty() ? TRect(GetParent()->GetSize()) : iTextListMenuProperties.iArea;
	SetRect(rect);
	SetTextStyle(iTextListMenuProperties.iTextStyle);
}

void TextListMenu::Construct(const Menu& aMenu)
{
	SetProperties();

	int halfItems = iNoOfMenuItems/2;
	for (int i=0 ; i<iNoOfMenuItems ; i++)
	{
		int deltaFromMiddle = (i <= halfItems) ? (halfItems - i) : (i - halfItems);

		MenuItemWidget* menuItemWidget = ConstructMenuItemWidget(deltaFromMiddle);
		if (menuItemWidget)
		{
			menuItemWidget->SetRect(GetItemArea(i));
			AddWidget(*menuItemWidget, true); // transfer ownership
			iMenuItemWidgets.push_back(menuItemWidget);
		}
		else
		{
			printf("TextListMenu::Construct: oops, ConstructMenuItemWidget(%d) failed\n", i);
		}
	}
	
	Populate(aMenu);
}

void TextListMenu::SetMenuItem(int aIndex, const MenuItem& aMenuItem)
{
	if (aIndex>=0 && aIndex<(int)iMenuItemWidgets.size())
	{
		bool isSubMenu = aMenuItem.GetSubMenu() != nullptr;
		iMenuItemWidgets[aIndex]->SetMenuText(aMenuItem.GetName(), isSubMenu);
	}
	else
	{
		printf("OOPS: TextListMenu::SetMenuItem index=%d not in range (0, %d)\n", aIndex, (int) iMenuItemWidgets.size());
	}
}


MenuItemWidget* TextListMenu::ConstructMenuItemWidget(int aDeltaFromMiddle)
{
	int fontSize = iTextListMenuProperties.iFontSizeSelected;
	if (aDeltaFromMiddle != 0)
	{
		fontSize = (iTextListMenuProperties.iFont.iSize > 0) ? iTextListMenuProperties.iFont.iSize: SDLThemeProperties.iDefaultFont.iSize;
		fontSize -= (aDeltaFromMiddle * iTextListMenuProperties.iFontSizeGradient);
	}

	TColor color(iTextListMenuProperties.iFont.iTextColor);
	color.Gradient(iTextListMenuProperties.iFontColorGradient * aDeltaFromMiddle);

	MenuItemWidget* menuItemWidget = new MenuItemWidget();
	
	if (menuItemWidget)
	{
		menuItemWidget->iLabelWidget.SetMargin(TMargin(5));
		menuItemWidget->iLabelWidget.SetFont(iTextListMenuProperties.iFont.iName.c_str(), fontSize);
		menuItemWidget->iLabelWidget.SetTextColor(color);
		menuItemWidget->iLabelWidget.SetAlignment(iHorizontalAlignement);
		menuItemWidget->iLabelWidget.SetAlignment(Widget::TVerticalAlignment::ECenter);
		menuItemWidget->iLabelWidget.SetEclipseText((aDeltaFromMiddle == 0) ? true : false);
		menuItemWidget->iLabelWidget.SetName(std::string("menu2dwidget::menuItemWidget::LabelWidget-") + std::to_string(iWidgets.size()));
		
		menuItemWidget->iImageWidget.SetSize(fontSize, fontSize);
		menuItemWidget->iImageWidget.SetName(std::string("menu2dwidget::menuItemWidget::ImageWidget-") + std::to_string(iWidgets.size()));
		
		menuItemWidget->SetShowDebugOutline();
		menuItemWidget->SetName(std::string("menu2dwidget-menuItemWidget: ") + std::to_string(iWidgets.size()));
	}
	
	return menuItemWidget;
}

TRect TextListMenu::GetItemArea(int aItemIndex, int aDeltaY) const
{
	int itemHeight = ItemHeight();
	int halfItems = iTextListMenuProperties.iItems/2;
	int deltaFromMiddle = (aItemIndex <= halfItems) ? (halfItems - aItemIndex) : (aItemIndex - halfItems);
	
	int y = (aItemIndex * itemHeight) + aDeltaY;
	int x = 0;
	if (iTextStyle == TTextStyle::ERound)
		x = CalcPositiononRoundCurve(y, deltaFromMiddle == 0);
	
	return TRect(TPoint(x, y), TSize(GetWidth() - x, ItemHeight()));
}


int TextListMenu::CalcPositiononRoundCurve(int aItemTop, bool aIsMiddleItem) const
{
	int itemHeight = ItemHeight();
	int halfRoundLength = GetHeight() / 2;
	
	int x_OnRoundCurve = (aItemTop < halfRoundLength) ? aItemTop - halfRoundLength  :  halfRoundLength - aItemTop - itemHeight;
	int y_OnRoundCurve = iTextListMenuProperties.iRound - ((iTextListMenuProperties.iRound * x_OnRoundCurve * x_OnRoundCurve)/(halfRoundLength*halfRoundLength));

	if (aIsMiddleItem)
		y_OnRoundCurve = iTextListMenuProperties.iRound;

	return y_OnRoundCurve;
}

void TextListMenu::VerticalAnimate(int aStep, bool aIsAnimatingTopReached)
{
	int i=0;
	for (auto &widget : iMenuItemWidgets)
	{
		TPoint point;
		int deltaY = ItemHeight() - aStep;
		if (!aIsAnimatingTopReached)
			deltaY *= -1;
		
		widget->SetPosition(GetItemTopLeft(i, deltaY));
		
		i++;
	}
}


void TextListMenu::CreateCachedTextures(IRenderer& aRenderer)
{
	if (iTextListMenuProperties.iUseHighlight)
	{
		TRect rect = GetItemArea(iTextListMenuProperties.iItems/2);
		ITexture* texture = RenderUtils::CreateSVGRoundedRectTexture(aRenderer, rect.Size(), iTextListMenuProperties.iHighlightRoundedRect);
		if (texture)
			CacheAdd(*texture, rect.Position());
	}
}


/*********************************************
 * 
 * MenuImageListContainer
 * 
 *********************************************/

void ImageListMenu::Construct(const Menu& aMenu)
{
	// construct images and add to iImages lists - this is in menu order
	int halfItems = iImageListMenuProperties.iItems/2;
	const TSize size = iImageListMenuProperties.iImageSize;
	const TSize smallSize(size.iWidth * iImageListMenuProperties.iStackedImageRatio	, size.iHeight * iImageListMenuProperties.iStackedImageRatio);
	const TPoint topLeftSelectedItem(iImageListMenuProperties.iSelectedItemX, (GetSize().iHeight/2) - (size.iHeight/2));
	const TPoint bottomLeftSelectedItem(iImageListMenuProperties.iSelectedItemX, (GetSize().iHeight/2) + (size.iHeight/2));
	for (int i=0 ; i<iImageListMenuProperties.iItems ; i++)
	{
		ImageWidget* image = new ImageWidget();
		if (image)
		{

			int deltaFromMiddle = (i <= halfItems) ? (halfItems - i) : (i - halfItems);
			if (deltaFromMiddle != 0)
			{
				int delta = iImageListMenuProperties.iStackWidth * (deltaFromMiddle - 1);
				TPoint p = (i<halfItems) ?
					topLeftSelectedItem + iImageListMenuProperties.iPreviousItemTopLeftOffset - TPoint(delta)
					:
					bottomLeftSelectedItem + iImageListMenuProperties.iNextItemTopLeftOffset - TPoint(delta, -delta);

				image->SetRect(TRect(p, smallSize));
				image->SetDrawMode(ImageWidget::TDrawMode::EScaled);
			}
			else
				image->SetRect(TRect(topLeftSelectedItem, size));
			
			image->SetShowDebugOutline();
			iImages.push_back(image);
		}
		else
		{
			printf("ImageListMenu::Construct: could not create ImageWidget(%d)\n", i);
		}
	}

	Populate(aMenu);
}


void ImageListMenu::Render(IRenderer& aRenderer)
{
	int halfItems = iImageListMenuProperties.iItems/2;

	for (int i=0 ; i<halfItems ; i++)
		iImages[i]->Render(aRenderer);
	for (int i=iNoOfMenuItems-1 ; i>halfItems ; i--)
		iImages[i]->Render(aRenderer);
	iImages[halfItems]->Render(aRenderer);
}


void ImageListMenu::SetMenuItem(int aIndex, const MenuItem& aMenuItem)
{
	iImages[aIndex]->SetImageFile(iImageProvider.GetFileName(aMenuItem, iImageListMenuProperties.iLogicalImageDirectory));
}

void ImageListMenu::VerticalAnimate(int aStep, bool aIsAnimatingTopReached)
{
	int i=0;
	for (auto &image : iImages)
	{
		TPoint delta(0, 0);
		int deltaY = (aStep == 2) ? GetVerticalAnimateMaxSteps() - 1 : -1;
		int deltaX = -deltaY;
		
		if (!aIsAnimatingTopReached)
			deltaY = -deltaY;
		delta = TPoint(deltaX, deltaY*4);

		image->SetPosition(image->GetPosition() + delta);
		
		i++;
	}
}

TPoint ImageListMenu::GetMidPoint(int aItem)
{
	TPoint midPoint;
#if 0
	int halfItems = iImageListMenuProperties.iItems/2;
	const TSize size = iImageListMenuProperties.iImageSize;
	const TPoint topLeftSelectedItem(iImageListMenuProperties.iSelectedItemX, (GetSize().iHeight/2) - (size.iHeight/2));
	const TPoint bottomLeftSelectedItem(iImageListMenuProperties.iSelectedItemX, (GetSize().iHeight/2) + (size.iHeight/2));
	
	int deltaFromMiddle = (aItem <= halfItems) ? (halfItems - aItem) : (aItem - halfItems);
	if (deltaFromMiddle != 0)
	{
		int delta = iImageListMenuProperties.iStackWidth * (deltaFromMiddle - 1);
		midPoint = (aItem<halfItems) ?
		topLeftSelectedItem + iImageListMenuProperties.iPreviousItemTopLeftOffset - TPoint(delta)
		:
		bottomLeftSelectedItem + iImageListMenuProperties.iNextItemTopLeftOffset - TPoint(delta, -delta);
	}
	return p;
#endif
	return midPoint;
}

void ImageListMenu::VerticalScrollAnimate(int aStep, bool aAnimateUp)
{
#if 0
	int i=0;
	for (auto &image : iImages)
	{
		TPoint delta(0, 0);
		int deltaY = (aStep == 2) ? GetVerticalAnimateMaxSteps() - 1 : -1;
		int deltaX = -deltaY;
		
		if (!aIsAnimatingTopReached)
			deltaY = -deltaY;
		delta = TPoint(deltaX, deltaY*4);
		
		image->SetPosition(image->GetPosition() + delta);
		
		i++;
	}
#endif
}


/*********************************************
 * 
 * Menu2DWidget
 * 
 *********************************************/

void Menu2DWidget::SetDefaultProperties()
{
	iMenu2DProperties = SDLThemeProperties.iDefaultMenu2DUIProperties;
	iMenu2DProperties.iTextListMenuProperties.iFont = SDLThemeProperties.iDefaultFont;
	ApplyProperties();
}

void Menu2DWidget::ApplyProperties()
{
	SetRect(iMenu2DProperties.iRect);	
}

void Menu2DWidget::Construct()
{
	SetName("Menu2DWidget");
	ApplyProperties();
	SetShowDebugOutline();
	iCurrentMenuItemContainer = ConstructMenuItemWidgets(GetCurrentMenu());
}

bool Menu2DWidget::IsImageMenu(const Menu& aMenu)
{
	return std::find(iImageMenuTypes.begin(), iImageMenuTypes.end(), aMenu.GetType()) != iImageMenuTypes.end();
}

MenuItemContainer* Menu2DWidget::ConstructMenuItemWidgets(const Menu& aMenu)
{
	MenuItemContainer* menuItemContainer =	(IsImageMenu(aMenu) && iMenu2DProperties.iEnableImages) ?
		(MenuItemContainer*) new ImageListMenu(iMenu2DProperties.iImageListMenuProperties, iImageProvider)
		:
		(MenuItemContainer*) new TextListMenu(iMenu2DProperties.iTextListMenuProperties);
	
	if (menuItemContainer)
	{
		AddWidget(*menuItemContainer, true); // transfer owenership
		menuItemContainer->SetSize(GetSize()); // set to my size
		menuItemContainer->Construct(aMenu);
	}
	
	return menuItemContainer;
}


void Menu2DWidget::PopulateMenu(const Menu& aMenu)
{
	if (iCurrentMenuItemContainer)
		iCurrentMenuItemContainer->Populate(aMenu);
}


void Menu2DWidget::StartVerticalAnimating(bool aTopReached)
{
	if (iMenu2DProperties.iVerticalAnimation && !IsAnimating())
		iVerticalAnimation.StartAnimating(iCurrentMenuItemContainer, aTopReached, iMenu2DProperties.iVerticalAnimationDelay);
}


void Menu2DWidget::StartHorizontalAnimating(bool aScrollLeft)
{
	if (iMenu2DProperties.iHorizontalAnimation)
	{
		if (iNewMenuItemContainer && iCurrentMenuItemContainer)
			iHorizontalAnimation.StartAnimating(*iCurrentMenuItemContainer, *iNewMenuItemContainer, aScrollLeft, iMenu2DProperties.iHorizontalAnimationDelay);
	}
	else
		StopHorizontalAnimating(); // make sure the old list is removed
}

void Menu2DWidget::StopHorizontalAnimating()
{
	if (iNewMenuItemContainer)
	{
		if (iCurrentMenuItemContainer)
			RemoveWidget(*iCurrentMenuItemContainer);
		iCurrentMenuItemContainer = iNewMenuItemContainer;
		iNewMenuItemContainer = nullptr;
	}
	
	iHorizontalAnimation.StopAnimating();
}


bool Menu2DWidget::ScrollDown()
{
	bool wantRender = false;
	Menu& currentMenu = iMenuModel.CurrentMenu();

	if (!currentMenu.IsAtBottom())
	{
		UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuVerticalScroll);

		wantRender = true;
		
		if (currentMenu.ScrollDown() && currentMenu.MenuSize()>0)
			StartVerticalAnimating(false);

		PopulateMenu(currentMenu);
	}
	else
		UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuEndList);

	return wantRender;
}


bool Menu2DWidget::ScrollUp()
{
	bool wantRender = false;
	Menu& currentMenu = iMenuModel.CurrentMenu();
	
	if (!currentMenu.IsAtTop())
	{
		UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuVerticalScroll);
		
		wantRender = true;
		
		if (currentMenu.ScrollUp() && currentMenu.MenuSize()>0)
			StartVerticalAnimating(true);

		PopulateMenu(currentMenu);
	}
	else
		UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuEndList);
	
	return wantRender;
}

bool Menu2DWidget::ScrollPageDown()
{
	bool wantRender = false;
	Menu& currentMenu = iMenuModel.CurrentMenu();
	
	if (!currentMenu.IsAtBottom())
	{
		UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuVerticalScroll);
		
		wantRender = true;
		
		if (currentMenu.ScrollPageDown() && currentMenu.MenuSize()>0)
			StartVerticalAnimating(false);
		
		PopulateMenu(currentMenu);
	}
	else
		UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuEndList);
	
	return wantRender;
}


bool Menu2DWidget::ScrollPageUp()
{
	bool wantRender = false;
	Menu& currentMenu = iMenuModel.CurrentMenu();
	
	if (!currentMenu.IsAtTop())
	{
		UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuVerticalScroll);
		
		wantRender = true;
		
		if (currentMenu.ScrollPageUp() && currentMenu.MenuSize()>0)
			StartVerticalAnimating(true);
		
		PopulateMenu(currentMenu);
	}
	else
		UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuEndList);
	
	return wantRender;
}

void Menu2DWidget::SwitchToMenu(Menu& aMenu, bool aScrollLeft)
{
	UserFeedbackManager::SendUserFeedbackEvent(UserFeedbackManager::TEvent::EMenuHorizontalScroll);
	
	iNewMenuItemContainer = ConstructMenuItemWidgets(aMenu);
	StartHorizontalAnimating(aScrollLeft);
}

bool Menu2DWidget::ScrollHorizontal(bool aScrollLeft)
{
	bool wantRender = false;
	
	if (!IsAnimating())
	{
		bool success = aScrollLeft ? iMenuModel.ActivatePrevMenu() : iMenuModel.ActivateNextMenu();
		if (success)
		{
			SwitchToMenu(GetCurrentMenu(), aScrollLeft);
			wantRender = true;
		}
	}
	
	return wantRender;
}


bool Menu2DWidget::GoBack()
{
	bool consumed = iMenuModel.GoBack();

	if (consumed)
		PopulateMenu(GetCurrentMenu());

	return consumed;
}


bool Menu2DWidget::HandleCommand(TCommand aCommand)
{
	bool consumed = false;
	
	switch ((int) aCommand)
	{
		case TCommand::ELeft:
		{
			if (!EventManager::IsLastCommandRepeated())
			{
				consumed = GoBack();
				if (!consumed)
					consumed = ScrollLeft();
			}
			break;
		}
		
		case TCommand::ERight:
		{
			if (!EventManager::IsLastCommandRepeated())
				consumed = ScrollRight();
			break;
		}
		
		case TCommand::EUp:
		{
			consumed = ScrollUp();
			break;
		}
		
		case TCommand::EDown:
		{
			consumed = ScrollDown();
			break;
		}
		
		case TCommand::EPageUp:
		{
			consumed = ScrollPageUp();
			break;
		}
		
		case TCommand::EPageDown:
		{
			consumed = ScrollPageDown();
			break;
		}

		case TCommand::ESelect:
		{
			consumed = iMenuModel.SelectItem();
			if (!consumed)
			{
				GetCurrentMenu().MenuItemCommand(aCommand);
				consumed = true;
			}
			PopulateMenu(GetCurrentMenu());
			break;
		}
		
		case TCommand::EBack:
		{
			consumed = GoBack();
			break;
		}
		case TCommand::EDelete:
		{
			GetCurrentMenu().MenuItemCommand(aCommand);
			consumed = true;
			PopulateMenu(GetCurrentMenu());
			break;
		}
		
		default:
		{
			if (aCommand >= TCommand::EUserDefined)
			{
				GetCurrentMenu().MenuItemCommand(aCommand);
				consumed = true;
			}
			break;
		}
	}

	if (consumed)
		System::RequestReRender();
	
	return consumed;
}	

void Menu2DWidget::Render(IRenderer& aRenderer)
{
	// the horizontal animation is positioning widgets outside our widget's area so
	// we have to explicitly clip rendering to our widget's area
	RenderClipper clippper(aRenderer, GetScreenRect());
	
	bool wasHorAnimating = iHorizontalAnimation.IsAnimating();
	
	if (iVerticalAnimation.IsAnimating())
		iVerticalAnimation.Animate();
	if (iHorizontalAnimation.IsAnimating())
		iHorizontalAnimation.Animate();
	
	// a bit of a kludge but we need to cleanup when stopping horizontal animation
	if (wasHorAnimating && !iHorizontalAnimation.IsAnimating())
		StopHorizontalAnimating();
	
	Container::Render(aRenderer);
}

/*********************************************
 * 
 * Menu2DWidget::TVerticalAnimation
 * 
 *********************************************/

void Menu2DWidget::TVerticalAnimation::StartAnimating(MenuItemContainer* aMenuItemContainer, bool aTopReached, int aDelay)
{
	StopAnimating();
	
	iMenuItemContainer = aMenuItemContainer;
	iIsAnimatingTopReached = aTopReached;
	iDelay = aDelay; 
	
	iStep = 1;
	iMaxSteps = iMenuItemContainer ? iMenuItemContainer->GetVerticalAnimateMaxSteps() : 0;
}

void Menu2DWidget::TVerticalAnimation::Animate()
{
	iStep++;

	if (iMenuItemContainer)
		iMenuItemContainer->VerticalAnimate(iStep, iIsAnimatingTopReached);

	if (iDelay > 0)
		SDL_Delay(iDelay);

	if (iStep >= iMaxSteps + 1)
		StopAnimating();
	else
		System::RequestReRender();
}

/*********************************************
 * 
 * Menu2DWidget::THorizontalAnimation
 * 
 *********************************************/

void Menu2DWidget::THorizontalAnimation::StartAnimating(Widget& aCurrentListWidget, Widget& aNewListWidget, bool aAnimateLeft, int aDelay)
{
	StopAnimating();
	
	aNewListWidget.SetPosition(aNewListWidget.GetPosition() + TPoint(aNewListWidget.GetWidth() * (aAnimateLeft ? -1 : 1), 0));

	iCurrentListWidget = &aCurrentListWidget;
	iNewListWidget = &aNewListWidget;
	iStep = 1; 
	iStartPositionCurrentWidget = aCurrentListWidget.GetPosition();
	iStartPositionNewWidget = aNewListWidget.GetPosition();
	iDelta = ((float) aCurrentListWidget.GetWidth()) /  ((float) iMaxSteps);
	if (!aAnimateLeft)
		iDelta = -iDelta;
	iDelay = aDelay;
}

void Menu2DWidget::THorizontalAnimation::Animate()
{
	if (iCurrentListWidget && iNewListWidget)
	{
		float totalDelta = ((float) iStep) * iDelta;
		TPoint delta(totalDelta, 0);
		iCurrentListWidget->SetPosition(iStartPositionCurrentWidget + delta);
		iNewListWidget->SetPosition(iStartPositionNewWidget + delta);

		if (iDelay > 0)
			SDL_Delay(iDelay);
		
		iStep++;
		if (iStep >= iMaxSteps + 1)
			StopAnimating();

		System::RequestReRender();
	}
}


void Menu2DWidget::SwitchToMenu(const std::string& aMenuName)
{
	Menu* menu = iMenuModel.ActivateMenu(aMenuName);
	if (menu)
		SwitchToMenu(*menu);
}
