/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlscrollablecontainer.h"
#include "sdllayout.h"
#include "sdlthemeproperties.h"

using namespace sdl;

/*********************************************
 * 
 * ScrollableContainer
 * 
 *********************************************/

TRect ScrollableContainer::GetClientArea() const
{
	TRect rect = TRect(GetSize());
	if (iScrollbar.IsEnabled())
		rect.SetWidth(rect.Width() - iScrollbar.GetSize().iWidth);
	return rect;
}


void ScrollableContainer::ReLayoutIfNeeded()
{
	if (NeedsScrollbar() && !iScrollbar.IsEnabled())
	{
		iScrollbar.Enable();
		Layout(); // recursion 
	}
}

void ScrollableContainer::ResizeForFixedHeightItems()
{
	if(iLineItemModel == TLineItemModel::EFixedHeight)
	{
		if (iWidgets.size()> 0)
		{
			TSize	clientArea = GetClientAreaSize();
			int		itemHeight = iWidgets[0]->GetMinimumHeight();
			int itemsThatFit = GetHeight() / itemHeight;
			clientArea.iHeight = itemsThatFit * itemHeight;
		
			ResizeToClientArea(clientArea);
		}
	}
}

void ScrollableContainer::Layout()
{
	ResizeForFixedHeightItems();
	
	PreLayout();
	LayoutManager::ListLayout(*this);
	PostLayout();
	ReLayoutIfNeeded();
}

bool ScrollableContainer::NeedsScrollbar()
{
	bool NeedsScrollbar = false;
	if (iWidgets.size() > 1)
	{
		int totalHeight = GetLineHeight(0, GetNumberOfLines());
		NeedsScrollbar = totalHeight > GetClientAreaSize().iHeight;
	}
	return NeedsScrollbar;
}

void ScrollableContainer::PostLayout()
{
	iOriginalPosition = GetRect().Position();
	SetTopLine(0);
	if (iScrollbar.IsEnabled())
	{
		iScrollbar.SetModelHeight(GetLineHeight(0, GetNumberOfLines()));
		UpdateScrollbar();
	}
}

void ScrollableContainer::PositionScrollbar()
{
	TPoint p = GetScreenPosition(iOriginalPosition);
	p += TPoint(GetSize().iWidth - iScrollbar.GetWidth(), 0);
	iScrollbar.SetPosition(p);
}

void ScrollableContainer::SizeChanged(const TSize& aOldSize)
{
	iScrollbar.SetHeight(GetSize().iHeight);
	PositionScrollbar();
	Widget::SizeChanged(aOldSize);
}


int ScrollableContainer::GetLineHeight(int aStartLine, int aNumOfLines) const
{
	int totalHeight = 0;
	std::for_each(iWidgets.begin()+aStartLine, iWidgets.begin()+aStartLine+aNumOfLines, [&] (Widget* aWidget) {
		totalHeight += aWidget->GetSize().iHeight; } );
	return totalHeight;
}

void ScrollableContainer::SetContainerPosition()
{
	int y = GetLineHeight(0, iViewPortStartLine);
	SetPosition(TPoint(iOriginalPosition.iX, iOriginalPosition.iY - y));
}


void ScrollableContainer::SetBottomLine(int aLine)
{
	if (aLine>=0)
	{
		iViewPortStartLine = aLine;
		iViewPortNumberOfLines = 0;
		
		int height = 0;
		while(ViewPortAddLine(aLine--, height))
		{
			if (iViewPortNumberOfLines > 0)
				iViewPortStartLine--;
			iViewPortNumberOfLines++;
		}
	}
	else
	{
		iViewPortStartLine = 0;
		iViewPortNumberOfLines = 0;
	}
	
	SetContainerPosition();
}


void ScrollableContainer::SetTopLine(int aLine)
{
	iViewPortStartLine = aLine;
	iViewPortNumberOfLines = 0;
	
	int height = 0;
	while(ViewPortAddLine(aLine++, height))
		iViewPortNumberOfLines++;
	
	SetContainerPosition();
}

bool ScrollableContainer::ViewPortAddLine(int aLine, int& aCurrentheight)
{
	bool success = false;
	
	if (aLine >= 0 && aLine < GetNumberOfLines())
	{
		int height = GetLineHeight(aLine);
		int maxHeight = GetClientAreaSize().iHeight;
		if (height + aCurrentheight <= maxHeight)
		{
			aCurrentheight += height;
			success = true;
		}
	}
	return success;
}

bool ScrollableContainer::ViewPortScrollUp()
{
	bool success = false;
	
	if (!IsAtTop())
	{
		SetTopLine(iViewPortStartLine -1);
		UpdateScrollbar();
		success = true;
	}
	
	return success;
}


bool ScrollableContainer::ViewPortScrollDown()
{
	bool success = false;
	
	if (!IsAtBottom())
	{
		SetTopLine(iViewPortStartLine + 1);
		UpdateScrollbar();
		success = true;
	}
	
	return success;
}

void ScrollableContainer::ViewPortScrollToTop()
{
	SetTopLine(0);
	UpdateScrollbar();
}

void ScrollableContainer::ViewPortScrollToBottom()
{
	SetBottomLine(GetNumberOfLines()-1);
	UpdateScrollbar();
}

int ScrollableContainer::GetUnusedViewPortHeight()
{
	int renderedHeight = GetLineHeight(iViewPortStartLine, iViewPortNumberOfLines);
	return GetSize().iHeight - renderedHeight;
}

void ScrollableContainer::UpdateScrollbar()
{
	int top = -GetPosition().iY;
	if (IsAtBottom())
		top -= GetUnusedViewPortHeight();
	iScrollbar.SetModelTop(top);
}

void ScrollableContainer::Render(IRenderer& aRenderer)
{
	Widget::Render(aRenderer);
	for (int i = iViewPortStartLine ; i < iViewPortStartLine + iViewPortNumberOfLines ; i++)
	{
		if (i>=0 && i<(int)iWidgets.size())
			iWidgets[i]->Render(aRenderer);
		else
			printf("ScrollableContainer::Render, internal state error, i=%d, size%d\n", i, (int) iWidgets.size());
	}
	
	iScrollbar.Render(aRenderer);
}

void ScrollableContainer::HightlightChanged(THighlightChange aHighlightChange)
{
	switch (aHighlightChange)
	{
		case THighlightChange::EUp:
			if (iFocussedWidgetIndex+1 <= (GetViewPortStartLine() + GetViewPortNumberOfLines()/2))
				ViewPortScrollUp();
			break;
			
		case THighlightChange::EDown:
			if (iFocussedWidgetIndex >= GetViewPortStartLine() + GetViewPortNumberOfLines()/2)
				ViewPortScrollDown();
			break;
			
		case THighlightChange::ETop:
			ViewPortScrollToTop();
			break;

		case THighlightChange::EBottom:
			ViewPortScrollToBottom();
			break;
	}
}

bool ScrollableContainer::HandleCommandMyself(TCommand aCommand)
{
	bool consumed = Container::HandleCommandMyself(aCommand);
	
	if (!consumed)
	{
		switch (aCommand)
		{
			case TCommand::EDown:
				consumed = ViewPortScrollDown();
				break;
				
			case TCommand::EUp:
				consumed = ViewPortScrollUp();
				break;
				
			default:
			{
				consumed = false;
				break;
			}
		}
	}
	return consumed;
}
