/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlsimpletextwidget.h"
#include "sdlthemeproperties.h"

using namespace sdl;


static void GetFontFromCharFormat(const simpletext::TCharFormat& aCharFormat, TFont& aFont)
{
	aFont.iName = aCharFormat.iFontName.size() > 0 ? aCharFormat.iFontName : SDLThemeProperties.iDefaultFont.iName;
	aFont.iSize = (aCharFormat.iFontSize > 0 ) ? aCharFormat.iFontSize : SDLThemeProperties.iDefaultFont.iSize;
	aFont.iTextColor = aCharFormat.iTextColor;
	aFont.iBold = aCharFormat.iBold;
	aFont.iItalic = aCharFormat.iItalic;
	aFont.iUnderline = aCharFormat.iUnderline;
	aFont.iStrikeThrough = aCharFormat.iStrikethrough;
}


class SDLFontProvider : public simpletext::IFontProvider
{
public:
	SDLFontProvider() { }
	
public: // implements IFontProvider
	virtual int GetDescent(const simpletext::TCharFormat& aCharFormat) override
	{
		int descent = 0;
		TFont font;
		GetFontFromCharFormat(aCharFormat, font);
		descent = RenderUtils::GetDescent(font);
		
		return descent;
	}
	
	virtual bool FitText(const simpletext::TCharFormat& aCharFormat, const std::string& aText, std::string& aReturnText, int aMaxWidth, int& aWidth, int& aHeight, bool aEclipse) override
	{
		TFont font;
		GetFontFromCharFormat(aCharFormat, font);
		RenderUtils::FitText(font, aText, aReturnText, aMaxWidth, aWidth, aHeight, aEclipse);
		return true;
	}
	
	virtual void GetDefaultCharacterFormating(simpletext::TCharFormat& aCharFormat)  override
	{
		aCharFormat.Reset();
		aCharFormat.iFontSize = SDLThemeProperties.iDefaultFont.iSize;
		aCharFormat.iFontName = SDLThemeProperties.iDefaultFont.iName;
		aCharFormat.iTextColor = SDLThemeProperties.iDefaultFont.iTextColor.iColor;
		aCharFormat.iBold = SDLThemeProperties.iDefaultFont.iBold;
		aCharFormat.iItalic = SDLThemeProperties.iDefaultFont.iItalic;
		aCharFormat.iUnderline = SDLThemeProperties.iDefaultFont.iUnderline;
		aCharFormat.iStrikethrough = SDLThemeProperties.iDefaultFont.iStrikeThrough;
	}
};

void SimpleTextWidget::ApplyProperties()
{
	SetVerticalAlignment(iSimpleTextProperties.iVerticalAlign);
	if (iSimpleTextProperties.iSingleLine)
		LimitToOneLine(iSimpleTextProperties.iEclipseText);
}

void SimpleTextWidget::SetDefaultProperties()
{
	SetLineItemModel(ScrollableContainer::TLineItemModel::EVariableHeight);
	iSimpleTextProperties.iSingleLine = false;
	iSimpleTextProperties.iEclipseText = false;
	ApplyProperties();
}



void SimpleTextWidget::AddLineWidgets() 
{
	RemoveAllWidgets();
	for (int i=0 ; i < (int) iSimpleTextModel.GetLineInfos().size() ; i++)
	{
		Widget* widget = new SimpleTextLineWidget(i, iSimpleTextUI);
		AddWidget(*widget, true); // transfer ownership
	}
};

void SimpleTextWidget::PreLayout() 
{
	int width = GetSize().iWidth;
	if (width > 0)
	{
		SDLFontProvider fontProvider;
		iSimpleTextUI.Layout(width, fontProvider);
		AddLineWidgets();
	}
}

TRect SimpleTextWidget::GetClientArea() const
{
	TRect rect = ScrollableContainer::GetClientArea();
	rect += TPoint(0, GetVerticalAlignedY(GetTotalHeight(), rect.Height()));
	return rect;
}


SimpleTextWidget::SimpleTextLineWidget::SimpleTextLineWidget(int aLine, simpletext::View& aSimpleTextView) :
iLine(aLine), iSimpleTextView(aSimpleTextView)
{
}

void SimpleTextWidget::SimpleTextLineWidget::Render(IRenderer& aRenderer)
{
	TRect rect = GetScreenRect();
	iRenderer = &aRenderer;
	iSimpleTextView.RenderLine(*this, iLine, rect.X(), rect.Y(), rect.Width());
}

int SimpleTextWidget::SimpleTextLineWidget::GetMinimumHeight() const
{
	return iSimpleTextView.iModel.GetLineInfos()[iLine].iHeight;
}

int SimpleTextWidget::SimpleTextLineWidget::RenderWord(int aX, int aY, const std::string& aWord, const simpletext::TCharFormat& aCharFormat)
{
	TSize size;
	TFont font;
	GetFontFromCharFormat(aCharFormat, font);
	font.iTextColor = (aCharFormat.iTextColor == TColor::KUndefinedColor)? SDLThemeProperties.iDefaultFont.iTextColor : aCharFormat.iTextColor;
	if (iRenderer)
		size = RenderUtils::RenderText(*iRenderer, TPoint(aX, aY), font, aWord);
	
	return size.iWidth;
}

int SimpleTextWidget::GetTotalHeight() const
{
	int height = 0;
	for (int l=0 ; l<GetNumberOfLines() ; l++)
		height += GetLineHeight(l);
	
	return height;
}








