/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "sdlspinnerwidget.h"
#include "sdlthemeproperties.h"
#include "sdlpopupmenuwidget.h"

using namespace sdl;

float SpinnerWidget::sArrowRatio = 0.6f;

void SpinnerWidget::SpinnerModelList::SetCurrentString(const std::string& aString)
{
	iIndex = std::find(iList.begin(), iList.end(), aString) - iList.begin();
	if (iIndex < 0 || iIndex>(int)iList.size())
		iIndex = 0;
}


std::string SpinnerWidget::SpinnerModelList::ToString() const
{
	std::string value;
	int size = iList.size();
	if (size > 0)
	{
		int index = iIndex;
		if (index >= 0 && index<size)
			value = iList[index];
	}
	
	return value;
}

bool SpinnerWidget::SpinnerModelList::Next()
{
	bool changed = false;
	if (iIndex < (int)iList.size()-1)
	{
		iIndex++;
		changed = true;
	}

	return changed;
}


bool SpinnerWidget::SpinnerModelList::Previous()
{
	bool changed = false;
	if (iIndex > 1)
	{
		iIndex--;
		changed = true;
	}

	return changed;
}


bool SpinnerWidget::SpinnerModelList::GetValueList(std::vector<std::string>& aValueList)
{
	aValueList = iList;
	return iList.size() > 0;
}


int SpinnerWidget::SpinnerModelInt::ClipValue(int aValue)
{
	if (aValue < iMinValue)
		return iMinValue;
	if (aValue > iMaxValue)
		return iMaxValue;
	
	return aValue;
}


SpinnerWidget::SpinnerWidget(SpinnerModel& aSpinnerModel) : iSpinnerModel(aSpinnerModel)
{
	iLabel.SetAlignment(LabelWidget::THorizontalAlignment::ELeft);
	iLabel.SetAlignment(LabelWidget::TVerticalAlignment::ECenter);
	iLabel.SetMargin(TMargin(SDLThemeProperties.iDefaultFont.iSize/2, 0, SDLThemeProperties.iDefaultFont.iSize/2, 0));
	
	iLabelValue.SetAlignment(LabelWidget::THorizontalAlignment::ERight);
	iLabelValue.SetAlignment(LabelWidget::TVerticalAlignment::ECenter);
	iLabelValue.SetMargin(TMargin(SDLThemeProperties.iDefaultFont.iSize/2, 0, GetArrowSpaceSize().iWidth, 0));
	
	SetMaxWidthValueString(iSpinnerModel.GetMaxWidthSampleText());
}

void SpinnerWidget::SetMaxWidthValueString(const std::string& aSampleText)
{
	TFont font(SDLThemeProperties.iDefaultFont.iName, SDLThemeProperties.iDefaultFont.iSize);
	iMaxSpinnerWidth = RenderUtils::GetTextSize(iLabelValue.GetFont(), aSampleText).iWidth;
}

void SpinnerWidget::Render(IRenderer& aRenderer)
{
	Widget::Render(aRenderer);
	
	TRect screenRect = GetScreenRect();
	
	iLabel.SetTextColor(HasFocus() ? SDLThemeProperties.iDefaultHighlightTextColor : SDLThemeProperties.iDefaultFont.iTextColor);
	iLabelValue.SetTextColor(HasFocus() ? SDLThemeProperties.iDefaultHighlightTextColor : SDLThemeProperties.iDefaultFont.iTextColor);
	
	iLabel.SetPosition(screenRect.Position());
	iLabel.Render(aRenderer);
	
	iLabelValue.SetPosition(screenRect.Position());
	iLabelValue.Render(aRenderer);
	
	CreateArrowTextureIfNeeded(aRenderer);
	if (iArrowTexture.get() != nullptr)
	{
		{
			const TRect space(screenRect.GetTopRight() - TPoint(GetArrowSpaceSize().iWidth, 0), GetArrowSpaceSize());
			TRect rect = iArrowTexture.get()->GetRect();
			rect.Center(space);
			aRenderer.RenderTexture(*iArrowTexture.get(), rect.Position());
		}
		{
			int deltaW = (GetArrowSpaceSize().iWidth * 2) + iMaxSpinnerWidth;
			const TRect space(screenRect.GetTopRight() - TPoint(deltaW, 0), GetArrowSpaceSize());
			TRect rect = iArrowTexture.get()->GetRect();
			rect.Center(space);
			aRenderer.RenderTextureRotated(*iArrowTexture.get(), rect.Position(), 180);
		}
	}
}


void SpinnerWidget::Update()
{
	iLabelValue.SetText(iSpinnerModel.ToString());
}

void SpinnerWidget::Popup()
{
	std::vector<std::string> list;
	
	if (iSpinnerModel.GetValueList(list))
	{
		PopupMenuWidget popupMenu;
		popupMenu.SetPosition(GetScreenPosition());
		popupMenu.SetWidth(GetWidth());
		popupMenu.SetHeight(300);
		popupMenu.GetMenu().AddMenuItems(list);
		
		if (popupMenu.Execute())
		{
			iSpinnerModel.SetValueIndex(popupMenu.GetHighlightedItemIndex());
			Update();
		}
	}
}

bool SpinnerWidget::HandleCommand(TCommand aCommand)
{
	bool consumed = true;
	
	if (HasFocus())
	{
		switch(aCommand)
		{
			case  TCommand::ESelect:
			{
				Popup();
				break;
			}
			
			case  TCommand::ELeft:
			{
				iSpinnerModel.Previous();
				Update();
				break;
			}
			case  TCommand::ERight:
			{
				iSpinnerModel.Next();
				Update();
				break;
			}
			default:
				consumed = false;
				break;
		}
	}
	
	if (consumed)
		System::RequestReRender();
	
	return consumed;
}

void SpinnerWidget::CreateArrowTextureIfNeeded(IRenderer& aRenderer)
{
	if (iArrowTexture.get() == nullptr)
	{
		TColor frameColor = SDLThemeProperties.iControlProperties.iFrameColor;
		TColor fillColor  = SDLThemeProperties.iControlProperties.iFrameColor;
		iArrowTexture.reset(RenderUtils::CreateSVGArrowRightTexture(aRenderer, GetArrowSpaceSize().iHeight * .8, frameColor, fillColor));
	}
}

void SpinnerWidget::SizeChanged(const TSize& aOldSize)
{
	iLabel.SetSize(GetSize()); 
	iLabelValue.SetSize(GetSize()); 
	Widget::SizeChanged(aOldSize); 
}
