/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "simpletextmodel.h"
#include "simpletextparser.h"

#include <stdlib.h>
#include <string.h>
#include <map>

using namespace simpletext;

static int GetFontSize(const std::string& aFontSize, int aDefaultFontSize)
{
	int fontSize = atoi(aFontSize.c_str());
	if (fontSize<=0 || fontSize> 100)
	{
		if (aFontSize.compare("large") == 0)
			fontSize = aDefaultFontSize * 1.25;
		else if (aFontSize.compare("xlarge") == 0)
			fontSize = aDefaultFontSize * 1.5;
		else if (aFontSize.compare("xxlarge") == 0)
			fontSize = aDefaultFontSize * 2;
		else if (aFontSize.compare("small") == 0)
			fontSize = aDefaultFontSize - 2;
		else if (aFontSize.compare("xsmall") == 0)
			fontSize = aDefaultFontSize - 4;
		else if (aFontSize.compare("xxsmall") == 0)
			fontSize = aDefaultFontSize -6 ;
		else
			fontSize = aDefaultFontSize;
	}
	return fontSize;
}


class ParseConsumer  : public IParseConsumer
{
public:
	ParseConsumer(IFontProvider& aFontProvider, TLineInfos& alineInfos, int aMaxWidth, bool aLimitToOneLine, bool aEclipseLastWord) :
		iFontProvider(aFontProvider), iLineInfos(alineInfos), iMaxWidth(aMaxWidth), iLimitToOneLine(aLimitToOneLine), iEclipseLastWord(aEclipseLastWord)
	{
		iFontProvider.GetDefaultCharacterFormating(iDefaultCharFormat);
	}

	void ParseTagParagraph(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewParagraph();
		ParseTagAttributes(aTagName, aTagAttributes, iParseParagraphAttributesMap);
	}
	
	void ParseTagBold(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		iLineFragment.iCharFormat.iBold = true;
	}

	void ParseTagBoldOff(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		iLineFragment.iCharFormat.iBold = false;
	}

	void ParseTagItalic(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		iLineFragment.iCharFormat.iItalic = true;
	}

	void ParseTagItalicOff(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		iLineFragment.iCharFormat.iItalic = false;
	}

	void ParseTagUnderline(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		iLineFragment.iCharFormat.iUnderline = true;
	}
	
	void ParseTagUnderlineOff(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		iLineFragment.iCharFormat.iUnderline = false;
	}

	void ParseTagStrikethrough(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		iLineFragment.iCharFormat.iStrikethrough = true;
	}
	
	void ParseTagStrikethroughOff(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		iLineFragment.iCharFormat.iStrikethrough = false;
	}

	void ParseTagFont(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		NewCharFormat();
		ParseTagAttributes(aTagName, aTagAttributes, iParseFontAttributesMap);
		iLineFragment.iDescent = iFontProvider.GetDescent(iLineFragment.iCharFormat);
	}

	void ParseTagBreak(const std::string& aTagName, const TTagAttributes& aTagAttributes)
	{
		std::string emptyWord("");
		ConsumeWord(emptyWord);
	}

	virtual void ConsumeTag(const Tag& aTag) override
	{
		if (!IsConsumptionCompleted())
			DoConsumeTag(aTag);
	}

	void DoConsumeTag(const Tag& aTag)
	{
		auto it = iParseTagMap.find(aTag.iName);
		if (it != iParseTagMap.end())
		{
			(this->*it->second)(aTag.iName, aTag.iAtributes);
		}
		else
		{
			std::string s("Ignoring invalid tag: ");
			s.append(aTag.iName);
			FormatError(s);
		}
	}
	
	void ConsumeWord(std::string& aWord)
	{
		bool firstWord = iLineInfo.IsEmpty() && iLineFragment.IsEmpty();
		bool success = true;
		do
		{
			int maxWidth = iMaxWidth - iLineInfo.iWidth;
			if (!firstWord)
				maxWidth -= iLineFragment.iSpaceWidth; // add space between words. This could make maxWidth negative - this is ok, the next word simply won't fit
			
			int		width;
			int 	height;
			std::string partialWord;
			bool eclipseText = iLimitToOneLine && iEclipseLastWord;
			success = iFontProvider.FitText(iLineFragment.iCharFormat, aWord, partialWord, maxWidth, width, height, eclipseText);

			if (iLimitToOneLine && (aWord.size() != partialWord.size()))
			{
				// partial words for single-line models
				iLineFragment.iWords.push_back(partialWord);
				aWord.clear(); //done
				SetConsumptionCompleted(); // prevent consuming anything else
			}
			else if (firstWord || (aWord.size() == partialWord.size()))
			{
				iLineFragment.iWords.push_back(partialWord);
				aWord.erase(aWord.begin(), aWord.begin() + partialWord.size());
			}
			else
			{
				width = height = 0;
			}
		
			iLineInfo.iWidth += width;
			if (!firstWord)
				iLineInfo.iWidth += iLineFragment.iSpaceWidth; // add space between words
			if (iLineInfo.iHeight < height)
			{
				iLineInfo.iHeight = height;
				iLineInfo.iLineDescent = iLineFragment.iDescent;
			}

			firstWord = false;

			// aWord was too big - new line and fit more text
			if (aWord.size() > 0)
			{
				NewLine();
				firstWord = true;
			}
		} while ((aWord.size() > 0) && success);
	}
	
	static bool IsWWhiteSpace(char c) { return ((c==' ') || (c == '\t') | (c == '\n') || (c == '\r')); }
	
	static const char* SkipWhiteSpace(const char* aText, int& aLen)
	{
		while((aLen > 0) && IsWWhiteSpace(*aText))
		{
			aLen--;
			aText++;
		}
		return aText;
	}
	
	static const char* SkipNoneWhiteSpace(const char* aText, int& aLen)
	{
		while((aLen > 0) && !IsWWhiteSpace(*aText))
		{
			aLen--;
			aText++;
		}
		return aText;
	}
	
	static int GetWord(const char* aText, int aLen, std::string& aWord)
	{
		const char* wordStart = SkipWhiteSpace(aText, aLen);
		const char* wordEnd = SkipNoneWhiteSpace(wordStart, aLen);
		aWord.assign(wordStart, wordEnd-wordStart);
		
		return wordEnd - aText;
	}

	virtual void ConsumeText(const std::string& aText)
	{
		const char* text = aText.c_str();
		int len = aText.size();
		while (len > 0 && !IsConsumptionCompleted())
		{
			std::string word;
			int lenConsumed = GetWord(text, len, word);
			if (word.size() > 0)
			{
				ConsumeWord(word);
			}
			
			text += lenConsumed;
			len -= lenConsumed;
		}
	}
	
	virtual void Done() override
	{
		NewParagraph(); // make sure last fragment & line are added
	}
	
	virtual void Warning(int aStart, int aLen, const std::string& aError) override
	{
		printf("Parse error at position %d: %s\n", aStart, aError.c_str());
	}

private:
	void FormatError(const std::string& aError)
	{
		printf("Format error: %s\n", aError.c_str());
	}

	void ParseFontAttributeName(const std::string& aTagName, const std::string& aAttribute)
	{
		if (aAttribute.compare("default") == 0)
			iLineFragment.iCharFormat.iFontName = iDefaultCharFormat.iFontName;
		else
			iLineFragment.iCharFormat.iFontName.assign(aAttribute);
	}

	void ParseFontAttributeSize(const std::string& aTagName, const std::string& aAttribute)
	{
		iLineFragment.iCharFormat.iFontSize = GetFontSize(aAttribute, iDefaultCharFormat.iFontSize);
	}

	void ParseFontAttributeColor(const std::string& aTagName, const std::string& aAttribute)
	{
		if (aAttribute.compare("default") == 0)
		{
			iLineFragment.iCharFormat.iTextColor = iDefaultCharFormat.iTextColor;
		}
		else
		{
			size_t pos = aAttribute.find("0x");
			if (pos != std::string::npos)
			{
				char* p;
				iLineFragment.iCharFormat.iTextColor = strtol(aAttribute.c_str() + pos, &p, 16);
				if (*p != 0)
				{
					std::string s("tag <font> has invalid color attribute (invalid 0x hex value): ");
					s.append(aAttribute);
					FormatError(s);
				}
			}
			else
			{
				iLineFragment.iCharFormat.iTextColor = atoi(aAttribute.c_str());
			}
		}
	}
	
	void SetFont(const TTagAttributes& aTagAttributes)
	{
		for (auto &tagAttribute : aTagAttributes)
		{
			auto it = iParseFontAttributesMap.find(tagAttribute.first);
			if (it != iParseFontAttributesMap.end())
			{
				(this->*it->second)("", tagAttribute.second);
			}
			else
			{
				std::string s("tag <font> has invalid attribute key: ");
				s.append(tagAttribute.first);
				FormatError(s);
			}
		}
	}

	
	void NewCharFormat()
	{
		// only new character format if there is text - otherwise modify the existing one
		if (!iLineFragment.IsEmpty())
		{
			iLineInfo.iLineFragments.push_back(iLineFragment);
			iLineFragment.ResetText(); // leave iCharFormat unchanged
		}
	}
	
	void NewParagraph()
	{
		NewCharFormat();
		if (!iLineInfo.IsEmpty())
		{
			iLineInfos.push_back(iLineInfo);
			iLineInfo.ResetParagraph();
		}
		iLineFragment.Reset(); // force new char format
		iLineFragment.iCharFormat.Reset();
	}
	
	void NewLine()
	{
		if (!iLineFragment.IsEmpty()) 
			iLineInfo.iLineFragments.push_back(iLineFragment);
		iLineInfo.iIsLastLineInParagraph = false;
		iLineInfos.push_back(iLineInfo);
		iLineInfo.ResetLine();
		iLineFragment.ResetText(); // leave iCharFormat unchanged
	}
	
	void ParseParagraphAttributeAlign(const std::string& aTagName, const std::string& aAttribute)
	{
		if (aAttribute.compare("left") == 0)
			iLineInfo.iParaFormat.iAlign = TParaFormat::TAlignment::ELeft;
		else if (aAttribute.compare("right") == 0)
			iLineInfo.iParaFormat.iAlign = TParaFormat::TAlignment::ERight;
		else if (aAttribute.compare("center") == 0)
			iLineInfo.iParaFormat.iAlign = TParaFormat::TAlignment::ECenter;
		else if (aAttribute.compare("justified") == 0)
			iLineInfo.iParaFormat.iAlign = TParaFormat::TAlignment::EJustified;
		else
		{
			std::string s("tag <p> has align key with invalid attribute value: ");
			s.append(aAttribute);
			FormatError(s);
		}
	}

	bool IsConsumptionCompleted()
	{
		return (iLimitToOneLine && iConsumptionCompleted);
	}

	void SetConsumptionCompleted()
	{
		iConsumptionCompleted = true;
	}
	
	
	IFontProvider&		iFontProvider;
	TLineInfos& 		iLineInfos;
	TLineInfo			iLineInfo;
	TLineFragment		iLineFragment;
	int 				iMaxWidth;
	TCharFormat			iDefaultCharFormat;
	bool				iLimitToOneLine;
	bool				iEclipseLastWord;
	bool				iConsumptionCompleted = false;
	
	typedef void (ParseConsumer::*ParseTagMethod)(const std::string&, const TTagAttributes&);
	typedef std::map <std::string, ParseTagMethod> TParseTagMap;
	TParseTagMap iParseTagMap = 
	{
		{"p", &ParseConsumer::ParseTagParagraph},
		{"b", &ParseConsumer::ParseTagBold},
		{"/b", &ParseConsumer::ParseTagBoldOff},
		{"u", &ParseConsumer::ParseTagUnderline},
		{"/u", &ParseConsumer::ParseTagUnderlineOff},
		{"i", &ParseConsumer::ParseTagItalic},
		{"/i", &ParseConsumer::ParseTagItalicOff},
		{"s", &ParseConsumer::ParseTagStrikethrough},
		{"/s", &ParseConsumer::ParseTagStrikethroughOff},
		{"font", &ParseConsumer::ParseTagFont},
		{"br", &ParseConsumer::ParseTagBreak}
	};
	
	typedef void (ParseConsumer::*ParseTagAttribute)(const std::string&, const std::string&);
	typedef std::map <std::string, ParseTagAttribute> TParseAttributesMap;
	TParseAttributesMap iParseFontAttributesMap = 
	{
		{"name", &ParseConsumer::ParseFontAttributeName},
		{"size", &ParseConsumer::ParseFontAttributeSize},
		{"color", &ParseConsumer::ParseFontAttributeColor}
	};

	TParseAttributesMap iParseParagraphAttributesMap = 
	{
		{"align", &ParseConsumer::ParseParagraphAttributeAlign}
	};

	void ParseTagAttributes(const std::string& aTagName, const TTagAttributes& aTagAttributes, TParseAttributesMap& aParseAttributesMap)
	{
		for (auto &tagAttribute : aTagAttributes)
		{
			auto it = aParseAttributesMap.find(tagAttribute.first);
			if (it != aParseAttributesMap.end())
			{
				(this->*it->second)(aTagName, tagAttribute.second);
			}
			else
			{
				std::string s("tag <" + aTagName + "> has invalid attribute key: ");
				s.append(tagAttribute.first);
				FormatError(s);
			}
		}
	}
};


void Model::Layout(int aMaxWidth, IFontProvider& aFontProvider)
{
	iLineInfos.clear();
	ParseConsumer	parseConsumer(aFontProvider, iLineInfos, aMaxWidth, iLimitToOneLine, iEclipseLastWord);
	Parser			parser;
	
	parser.Parse(iText.c_str(), parseConsumer);
}

int Model::GetHeight() const
{
	int height = 0;
	for (auto &lineInfos : iLineInfos)
	{
		height += lineInfos.iHeight;
	}
	return height;
}

