/*******************************************************************
 * 
 *  Author: Peter van Sebille (peter@yipton.net)
 *
 *  (c) Copyright 2016, Peter van Sebille
 *  All Rights Reserved
 *
 *******************************************************************/

#include "vlcwidget.h"
#include "platform.h"
#include "sdleventmanager.h"
#include "vlc/vlc.h"
#include "iostream"

static char* sChromaFormat = "RGBA";

static bool sMuteSetting = false;
static int  sAudiolevelSetting = 50;

void VlcWidgetAudioControl::SetMute(bool aMute)
{
	sMuteSetting = aMute; 
}

bool VlcWidgetAudioControl::GetMute()
{
	return sMuteSetting; 
}

void VlcWidgetAudioControl::SetAudioLevel(int aAudioLevel)
{
	sAudiolevelSetting = aAudioLevel; 
}

int VlcWidgetAudioControl::GetAudioLevel()
{
	return sAudiolevelSetting; 
}



void* VlcWidget::LockCallback(void* aContext, void** aPixels) 
{
	VlcWidget* vlcWidget = (VlcWidget*)aContext;
	vlcWidget->Lock(aPixels);
	return nullptr;
}

void VlcWidget::UnlockCallback(void* aContext, void* /*aId*/, void* const* /*aPixels*/) 
{
	VlcWidget* vlcWidget = (VlcWidget*)aContext;
	vlcWidget->Unlock();
}

static bool wait_playing(libvlc_media_player_t* aMediaPlayer)
{
	int count = 0;
	libvlc_state_t state;
	do 
	{
		state = libvlc_media_player_get_state (aMediaPlayer);
		SDL_Delay(1);
	} while ((state != libvlc_Playing) && (state != libvlc_Error) && (state != libvlc_Ended) && (count++<1000));
	return state == libvlc_Playing;
}


class VlcManager
{
public:
	virtual ~VlcManager();
	VlcManager();
	
	libvlc_instance_t* Instance()			{ return iLibVlcInstance; }
private:
	void Construct();
	void Release();
	
	libvlc_instance_t* iLibVlcInstance = nullptr;
};

VlcManager::VlcManager()
{
	Construct();
}

void VlcManager::Construct()
{
	iLibVlcInstance = libvlc_new(0, nullptr);
	if(iLibVlcInstance == nullptr) 
		std::cout << "VlcManager: LibVLC initialization failure." << std::endl;
}

VlcManager::~VlcManager()
{
	Release();
}

void VlcManager::Release()
{
	if (iLibVlcInstance)
	{
		libvlc_release(iLibVlcInstance);
		iLibVlcInstance = nullptr;
	}
}

static VlcManager sVlcManager;


VlcWidget::VlcWidget()
{
	iMutex = SDL_CreateMutex();
	iTimeLabel.SetAlignment(sdl::Widget::THorizontalAlignment::ERight);
	iTimeLabel.SetAlignment(sdl::Widget::TVerticalAlignment::EBottom);
	iTimeLabel.SetFont("LiberationMono-Bold.ttf", 15);
	iTimeLabel.SetTextColor(0xffffff);
}

VlcWidget::~VlcWidget()
{
	Stop();

	if (iMutex)
	{
		SDL_DestroyMutex(iMutex);
		iMutex = nullptr;
	}
}

void VlcWidget::Lock(void** aPixels)
{
	SDL_LockMutex(iMutex);
	*aPixels = iVideoBuffer;
}

void VlcWidget::Unlock()
{
	SDL_UnlockMutex(iMutex);
}


void VlcWidget::SetupAudioVolume()
{
	int audioLevel = VlcWidgetAudioControl::GetAudioLevel();
	if (VlcWidgetAudioControl::GetMute())
		audioLevel = 0;
	if (audioLevel<0 || audioLevel>100)
		audioLevel = 0;
	
	libvlc_audio_set_volume((libvlc_media_player_t*)iVlcPlayer, audioLevel);
}


bool VlcWidget::PlayVideoFile(const std::string& aFileName)
{
	bool success = false;
	
	if (!Platform::FileExists(aFileName))
		return false;
	
	Stop();
	
	if (sVlcManager.Instance())
	{
		libvlc_media_t* vlcMedia = libvlc_media_new_path(sVlcManager.Instance(), aFileName.c_str());
		if (vlcMedia)
		{
			libvlc_media_parse(vlcMedia);
			libvlc_media_player_t* vlcPlayer = libvlc_media_player_new_from_media(vlcMedia);
			if (vlcPlayer)
			{
				iVideoDuration = (int) libvlc_media_get_duration(vlcMedia);
				iVideoStartTime = SDL_GetTicks();
				
				unsigned width=0, height=0;
				libvlc_video_get_size(vlcPlayer, 0, &width, &height);
				iVideoSize = sdl::TSize((int) width, (int) height);
				if (!iVideoSize.IsEmpty())
				{
					iVlcPlayer = vlcPlayer;
					libvlc_media_release(vlcMedia);
					vlcMedia = nullptr;

					libvlc_video_set_callbacks(vlcPlayer, LockCallback, UnlockCallback, nullptr, (void*) this);
					
					int pitch = iVideoSize.iWidth*4;
					libvlc_video_set_format(vlcPlayer, sChromaFormat, iVideoSize.iWidth, iVideoSize.iHeight, pitch);
					
					iVideoBufferSize = iVideoSize.iHeight * pitch;
					iVideoBuffer = (char*) malloc(iVideoBufferSize);
					memset(iVideoBuffer, 0, iVideoBufferSize);
					
					if (iVideoBuffer)
					{
						SetupAudioVolume();
						
						if (libvlc_media_player_play(vlcPlayer) == 0)
						{
							if (wait_playing(vlcPlayer))
								success = true;
							else
								printf("VlcWidget::PlayVideoFile, although started, video is not playing (%s)\n", aFileName.c_str());
						}
						else
							printf("VlcWidget::PlayVideoFile, libvlc_media_player_play failed (%s), error is: %s\n", aFileName.c_str(), libvlc_errmsg());
					}
					else
						printf("VlcWidget::PlayVideoFile, could not allocate video memory (%d) for video %s\n", iVideoBufferSize, aFileName.c_str());
				}
				else
					printf("VlcWidget::PlayVideoFile, could not determine video size (%s)\n", aFileName.c_str());
			}
			else
				printf("VlcWidget::PlayVideoFile, libvlc_media_player_new_from_media failed (%s), error: %s\n", aFileName.c_str(), libvlc_errmsg());
		}
		else
			printf("VlcWidget::PlayVideoFile, libvlc_media_new_path failed (%s), error: %s\n", aFileName.c_str(), libvlc_errmsg());
	}
	else
		printf("VlcWidget::PlayVideoFile, VLC is not supported\n");
	
	if (!success)
		Stop();

	return success;
}


void VlcWidget::CopyFrameToTexture()
{
	SDL_LockMutex(iMutex);
	
	if (iTexture)
		iTexture->CopyBytes(iVideoBuffer, iVideoSize.iWidth*4);
	
	SDL_UnlockMutex(iMutex);
}


void VlcWidget::CreateTextureIfNeeded(sdl::IRenderer& aRenderer)
{
	if (iVlcPlayer && (iTexture == nullptr))
		iTexture = aRenderer.CreateMemoryTexture(iVideoSize, sChromaFormat);
}

void VlcWidget::Render(sdl::IRenderer& aRenderer)
{
	CreateTextureIfNeeded(aRenderer);
	
	if (iTexture && iVideoBuffer && iVlcPlayer)
	{
		CopyFrameToTexture();
		sdl::TRect renderRect;
		sdl::RenderUtils::RenderScaledTextureMaintainAspectRatio(aRenderer, *iTexture, GetScreenRect(), renderRect);
		
		if (!IsPlaying())
			Stop();
		else
		{
			iTimeLabel.SetRect(renderRect + sdl::TSize(0, iTimeLabel.GetFontSize() * 1.5));
			//			iTimeLabel.SetRect(iRenderRect);
			SetTimeLabel();
			iTimeLabel.Render(aRenderer);
		}
		
		sdl::System::RequestReRender();
	}
}


void VlcWidget::Stop()
{
	if (iVlcPlayer)
	{
		libvlc_media_player_stop((libvlc_media_player_t*)iVlcPlayer);
		libvlc_media_player_release((libvlc_media_player_t*)iVlcPlayer);
		iVlcPlayer = nullptr;
	}

	if (iTexture)
	{
		delete(iTexture);
		iTexture = nullptr;
	}
	
	if (iVideoBuffer)
	{
		free(iVideoBuffer);
		iVideoBuffer = nullptr;
		iVideoBufferSize = 0;
	}
	
	iVideoSize = sdl::TSize(0, 0);
	
	iVideoDuration = 0;
	iVideoStartTime = 0;
}


bool VlcWidget::IsPlaying()
{
	return iVlcPlayer ? libvlc_media_player_is_playing((libvlc_media_player_t*)iVlcPlayer)==1 : false;
}

void VlcWidget::GetVideoTime(int& aDuration, int& aElapsedTime)
{
	if (IsPlaying())
	{
		aDuration = iVideoDuration;
		aElapsedTime = SDL_GetTicks() - iVideoStartTime;
	}
	else
	{
		aDuration = 0;
		aElapsedTime = 0;
	}
}

void VlcWidget::SetTimeLabel()
{
	int duration;
	int elapsedTime;
	GetVideoTime(duration, elapsedTime);

	char timeText[64];
	int durationMin = (duration/1000)/60;
	int durationSec = (duration/1000) % 60;
	int durationMsec = duration%1000;

	int elapsedMin = (elapsedTime/1000)/60;
	int elapsedSec = (elapsedTime/1000) % 60;
	int elapsedMsec = elapsedTime%1000;
	
	if (durationMin > 0)
		sprintf(timeText, "%d.%d.%03d:%d.%d.%03d", elapsedMin, elapsedSec, elapsedMsec, durationMin, durationSec, durationMsec);
	else
		sprintf(timeText, "%d.%03d:%d.%03d", elapsedSec, elapsedMsec, durationSec, durationMsec);
	iTimeLabel.SetText(timeText);
}

